UNIT KDataSet;
{+H
---------------------------------------------------------------------------
  File        - KDATASET.PAS
  Purpose     - Implements a multi-column, multi-data type descendant of TDataSet.
  Reference   - Based on TxtFile.pas by Vladimir Savelev; ugms@sakhmail.sakhalin.ru
  Copyright    1998 Keith S. Brown, except as noted above.
  Remarks     - Implements a multi-column, multi-data type dataset without using
                the BDE. Its primary purposes are:
                1. load & display data from a delimited text file, where the first
                   row is assumed to be the column headings.
                2. display static data stored in a form or exe.
                3. load & display computed data.

                As this component is a descedant of TDataSet, it can be used with
                most data aware controls, such as DBGrid, DBNavigator, etc.

                By default all fields are string fields. However, methods exist
                for changing the default on a column by column basis.

                With static data stored in a form or in the exe, the data can be
                changed... but the changes are not saved.

  Revised     - 1998.1019-23 (KSB) Added support for integer-type fields,
                float fields, date-type fields and currency fields.
              - 1998.1027 (KSB) Moved TDataStore to its own unit (KSSStore).
              - 1998.1117 (KSB) Added Rebookmark procedure;
              - 1998.1221 (KSB) Fixed data conversion error.
              - 1998.1222 (KSB) Rewrote FindKey method.
              - 1998.1223 (KSB) Rewrote GetColumnTypes method.
              - 1998.1230 (KSB) Added HasHeaderRow property.
---------------------------------------------------------------------------}
INTERFACE
USES
  Classes,  Db,       Consts,   SysUtils, KSSStore;

TYPE
  TRecInfo    = PACKED RECORD
    FiBookMark     : LongINT;
//  UpdateStatus   : TUpdateStatus;
    FeBookMarkFlag : TBookmarkFlag;
  END {RECORD};

  TFieldOffset= ARRAY[0..10000]OF INTEGER;
  PFieldOffset= ^TFieldOffset;
  TGetColNameEvent = PROCEDURE(aoSender:TObject; aiColumn,aiRows:WORD;
                      VAR asName  :STRING;
                          asValue :STRING;
                      VAR aeType  :TFieldType)OF OBJECT;
  PLongINT    = ^LongINT;
  PRecInfo    = ^TRecInfo;

              { TKSS_CustomDataSet - descedant of TDataSet,
                implements string, float, integer, date & currency fields}
  TKSS_CustomDataSet= CLASS(TDataSet)
  PRIVATE
    FaFldOffs      : PFieldOffset;
    FbByPassNIOpen : BOOLEAN;          {bypass normal internal open: Set in RefreshDisplayLabels}
    FbHasHdrRow    : BOOLEAN;
    FbKeepDataInMem: BOOLEAN;
    FbSaveChanges  : BOOLEAN;
    FiCurrentRec   : INTEGER;
    FiLastBookmark : LongINT;
    FiRecBufSize   : INTEGER;
    FiRecInfoOfs   : INTEGER;
    FiRecordSize   : INTEGER;
    FoAbout        : TObject;
    FoData         : TDataStore;
    FOnGetColName  : TGetColNameEvent;
    FUNCTION    GetDelimiter : CHAR;
    PROCEDURE   GetColumnTypes;
    PROCEDURE   ReadFld( aoReader: TReader );
    PROCEDURE   SetDelimiter( acDelimiter : CHAR );
    PROCEDURE   WriteFld( aoWriter: TWriter );
  PROTECTED
    FUNCTION    AllocRecordBuffer: PChar; OVERRIDE;
    PROCEDURE   DefineProperties(aoFiler: TFiler); OVERRIDE;
    PROCEDURE   FreeRecordBuffer(VAR apBuffer: PChar); OVERRIDE;
    PROCEDURE   GetBookmarkData(apBuffer: PChar; apData: POINTER); OVERRIDE;
    FUNCTION    GetBookmarkFlag(apBuffer: PChar): TBookmarkFlag; OVERRIDE;
    FUNCTION    GetFieldData(aoField: TField; apBuffer: POINTER): BOOLEAN; OVERRIDE;
    FUNCTION    GetRecNo: INTEGER; OVERRIDE;
    FUNCTION    GetRecord(apBuffer: PChar; aeGetMode: TGetMode; abDoCheck: BOOLEAN): TGetResult; OVERRIDE;
    FUNCTION    GetRecordCount: INTEGER; OVERRIDE;
    FUNCTION    GetRecordSize: WORD; OVERRIDE;
    PROCEDURE   InternalAddRecord(apBuffer: POINTER; abAppend: BOOLEAN); OVERRIDE;
    PROCEDURE   InternalClose; OVERRIDE;
    PROCEDURE   InternalDelete; OVERRIDE;
    PROCEDURE   InternalFirst; OVERRIDE;
    PROCEDURE   InternalGotoBookmark(apBookmark: POINTER); OVERRIDE;
    PROCEDURE   InternalHandleException; OVERRIDE;
    PROCEDURE   InternalInitFieldDefs; OVERRIDE;
    PROCEDURE   InternalInitRecord(apBuffer: PChar); OVERRIDE;
    PROCEDURE   InternalLast; OVERRIDE;
    PROCEDURE   InternalOpen; OVERRIDE;
    PROCEDURE   InternalPost; OVERRIDE;
    PROCEDURE   InternalSetToRecord(apBuffer: PChar); OVERRIDE;
    FUNCTION    IsCursorOpen: BOOLEAN; OVERRIDE;
    PROCEDURE   SetBookmarkData(apBuffer: PChar; apData: POINTER); OVERRIDE;
    PROCEDURE   SetBookmarkFlag(apBuffer: PChar; aeValue: TBookmarkFlag); OVERRIDE;
    PROCEDURE   SetFieldData(aoField: TField; apBuffer: POINTER); OVERRIDE;
    PROCEDURE   SetRecNo(aiValue: INTEGER); OVERRIDE;
  PROTECTED
    FUNCTION    GetHeader:TColumnDefs; VIRTUAL;
    PROCEDURE   LoadDataStore; VIRTUAL;
    PROCEDURE   RefreshDisplayLabels; VIRTUAL;
    PROCEDURE   ReportTo(asFileName:STRING; abAppend:BOOLEAN);
    PROCEDURE   SaveDataStore; VIRTUAL;
    PROCEDURE   SetHeader(aoHeader:TColumnDefs); VIRTUAL;
  PROTECTED
    PROCEDURE   KSS_BufferToStore(apBuffer:PChar);
    PROCEDURE   KSS_ResetRecordSize;
    FUNCTION    KSS_GetFieldSize(aiFieldNo:INTEGER):INTEGER;
    PROCEDURE   KSS_StoreToBuffer(apBuffer:PChar);
  PROTECTED
    PROPERTY    Data:TDataStore      READ FoData      WRITE FoData;
    PROPERTY    HasHeaderRow:BOOLEAN READ FbHasHdrRow WRITE FbHasHdrRow DEFAULT TRUE;
    PROPERTY    Header:TColumnDefs   READ GetHeader   WRITE SetHeader;
    PROPERTY    SaveChanges:BOOLEAN  READ FbSaveChanges;
  PUBLIC
    CONSTRUCTOR Create(aoOwner: TComponent); OVERRIDE;
    DESTRUCTOR  Destroy; OVERRIDE;
    FUNCTION    FindKey(CONST aaKeyValues: ARRAY OF CONST): BOOLEAN;
    PROCEDURE   Rebookmark;
    PROCEDURE   Report;
  PUBLIC
    PROPERTY    Delimiter: CHAR READ GetDelimiter WRITE SetDelimiter;
    PROPERTY    KeepDataInMemory : BOOLEAN READ FbKeepDataInMem WRITE FbKeepDataInMem;
  PUBLISHED
    PROPERTY    About : TObject READ FoAbout WRITE FoAbout;
    PROPERTY    OnGetColumnName :TGetColNameEvent READ FOnGetColName WRITE FOnGetColName;
  END {CLASS};

              { TKSS_DataSet - descedant of TKSS_CustomDataSet }
  TKSS_DataSet= CLASS(TKSS_CustomDataSet)
  PUBLIC
    PROPERTY    Data;
    PROPERTY    Header;
    PROPERTY    SaveChanges;
  PUBLISHED
    PROPERTY    AfterCancel;
    PROPERTY    AfterClose;
    PROPERTY    AfterDelete;
    PROPERTY    AfterEdit;
    PROPERTY    AfterInsert;
    PROPERTY    AfterOpen;
    PROPERTY    AfterPost;
    PROPERTY    AfterScroll;
    PROPERTY    BeforeCancel;
    PROPERTY    BeforeClose;
    PROPERTY    BeforeDelete;
    PROPERTY    BeforeEdit;
    PROPERTY    BeforeInsert;
    PROPERTY    BeforeOpen;
    PROPERTY    BeforePost;
    PROPERTY    BeforeScroll;
    PROPERTY    Delimiter;
    PROPERTY    KeepDataInMemory;
    PROPERTY    OnCalcFields;
    PROPERTY    OnDeleteError;
    PROPERTY    OnEditError;
    PROPERTY    OnNewRecord;
    PROPERTY    OnPostError;
  END {CLASS};

              { TTextTable - descedant of TKSS_DataSet,
                implements simple text file based dataset}
  TTextTable  = CLASS(TKSS_DataSet)
  PRIVATE
    FbBlankLinesOk : BOOLEAN;
    FsFileName     : STRING;
  PROTECTED
    PROCEDURE   LoadDataStore; OVERRIDE;
    PROCEDURE   SaveDataStore; OVERRIDE;
  PUBLISHED
    PROPERTY    Active;
    PROPERTY    AllowBlanks: BOOLEAN READ FbBlankLinesOk WRITE FbBlankLinesOk;
    PROPERTY    FileName:    STRING  READ FsFileName     WRITE FsFileName;
    PROPERTY    HasHeaderRow;
    PROPERTY    Header;
  END {CLASS};


{}FUNCTION    GetCopyright:STRING;
{}FUNCTION    GetVersion:STRING;
              {=================================================}
IMPLEMENTATION
USES
  Windows,  Forms,    KEnums,   KSSDates;


TYPE
  TBuffer     = ARRAY[0..10000]OF CHAR;
  PBuffer     =^TBuffer;
  TStorage    = CLASS(TDataStore);


{}FUNCTION    GetCopyright:STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Copyright string for About property.
---------------------------------------------------------------------------}
  BEGIN
    result := 'Freeware (Copyright  1998) By Keith Brown'#13
             +'TKSS_DataSet components'#13
             +'E-Mail: keith_brown@juno.com';
{}END {GetCopyright};




{}FUNCTION    GetVersion:STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Version string for About property.
---------------------------------------------------------------------------}
  BEGIN
    result := 'TKSS_DataSet components version 1.1.0';
{}END {GetVersion};


              {KSS Custom Data Set Methods}


{}FUNCTION    TKSS_CustomDataSet.AllocRecordBuffer: PChar;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Remarks     - TDataSet calls this method to allocate the record buffer.
                Here we use FiRecBufSize which is equal to the size of the
                data plus the size of the TRecInfo structure.
---------------------------------------------------------------------------}
  BEGIN
    GetMem(Result, FiRecBufSize);
{}END {AllocRecordBuffer};




{}CONSTRUCTOR TKSS_CustomDataSet.Create(aoOwner: TComponent);
{+O
---------------------------------------------------------------------------
  Purpose     - Allocate the data store and the list of column names.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Create(aoOwner);
    FaFldOffs       := NIL;
    FbBypassNIOpen  := FALSE;
    FbHasHdrRow     := TRUE;
    FbKeepDataInMem := TRUE;
    FbSaveChanges   := FALSE;
    FOnGetColName   := NIL;
    Data := TDataStore.Create;
{}END {Create};




{}PROCEDURE   TKSS_CustomDataSet.DefineProperties(aoFiler: TFiler);
{+O
---------------------------------------------------------------------------
  Purpose     - Define a FldNames property
---------------------------------------------------------------------------}
  BEGIN
    INHERITED DefineProperties(aoFiler);
    aoFiler.DefineProperty( 'FldNames', ReadFld, WriteFld, Data.Header.Count>0 );
    TStorage(Data).DefineProperties(aoFiler);
{}END {DefineProperties};




{}DESTRUCTOR  TKSS_CustomDataSet.Destroy;
{+O
---------------------------------------------------------------------------
  Purpose     - Release the data store and list of column names.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Destroy;
    Data.Free;
{}END {Destroy};




{}FUNCTION TKSS_CustomDataSet.FindKey(CONST aaKeyValues: ARRAY OF CONST): BOOLEAN;
{+O
---------------------------------------------------------------------------
  Purpose     - Find the row that matches the key.
  Revised     - 1998.1222 (KSB) Rewrote to handle NULL and multiple values
                in aaKeyValues.
---------------------------------------------------------------------------}
  CONST
    csBoolean : ARRAY[BOOLEAN]OF STRING[5]= ('FALSE','TRUE');
  VAR
    lbFound   : BOOLEAN;
    lbIsNull  : BOOLEAN;
    lbMatched : BOOLEAN;
    lbPartial : BOOLEAN;
    liCol     : INTEGER;
    liMaxCols : INTEGER;
    liRow     : INTEGER;
    liRowFound: INTEGER;
    lsConst   : STRING;
    lsData    : STRING;
  BEGIN
    Result := FALSE;
    liMaxCols := Data.ColCount-1;
    IF liMaxCols > High(aaKeyValues) THEN BEGIN
      liMaxCols := High(aaKeyValues);
    END {IF};

    FOR liRow := 0 TO Data.RowCount-1 DO BEGIN
      lbFound   := TRUE;
      FOR liCol := 0 TO liMaxCols DO BEGIN
        lbIsNull := FALSE;
        lbMatched:= FALSE;
        lsData   := Data.CellAsString[liCol,liRow];
        WITH aaKeyValues[liCol] DO BEGIN
          CASE VType OF
            vtInteger   : lsConst := IntToStr(VInteger);
            vtBoolean   : lsConst := csBoolean[VBoolean];
            vtChar      : lsConst := VChar;
            vtExtended  : lsConst := FloatToStr(VExtended^);
            vtString    : lsConst := VString^;
            vtPChar     : lsConst := VPChar;
            vtObject    : lsConst := VObject.ClassName;
            vtClass     : lsConst := VClass.ClassName;
            vtAnsiString: lsConst := STRING(VAnsiString);
            vtCurrency  : lsConst := CurrToStr(VCurrency^);
            vtVariant   : BEGIN
              lbIsNull:= VarIsNull(VVariant^);
              lsConst := STRING(VVariant^);
            END {IF};

            ELSE BEGIN
              lsConst := '';
            END {ELSE};
          END {CASE};

          lbPartial := VType IN [vtString,vtPChar,vtObject,vtClass];
          IF lbIsNull THEN BEGIN
            lbMatched := TRUE;
          END ELSE
          IF lbPartial AND (Copy(lsData,1,Length(lsConst)) = lsConst) THEN BEGIN
            lbMatched := TRUE;
          END ELSE
          IF lsData = lsConst THEN BEGIN
            lbMatched := TRUE;
          END {IF};

          lbFound := lbFound AND lbMatched;
          IF NOT lbMatched THEN BEGIN
            Break;
          END {IF};
        END {WITH};
      END {FOR};

      IF lbFound THEN BEGIN
        liRowFound := liRow;
        Break;
      END {IF};
    END {FOR};

    IF lbFound THEN BEGIN
      Result := TRUE;
      FiCurrentRec := liRowFound;
      Resync([]);
    END {IF};
{}END {FindKey};




{}PROCEDURE   TKSS_CustomDataSet.FreeRecordBuffer(VAR apBuffer: PChar);
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Remarks     - TDataSet calls this method to free the record buffer.
  Warning     - Make sure the value of FiRecBufSize does not change before
                all the allocated buffers are freed.
---------------------------------------------------------------------------}
  BEGIN
    FreeMem(apBuffer, FiRecBufSize);
{}END {FreeRecordBuffer};




{}PROCEDURE   TKSS_CustomDataSet.GetBookmarkData(apBuffer: PChar; apData: POINTER);
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Purpose     - Provides a way to read bookmark data from the record buffer
                without actually repositioning the current record
---------------------------------------------------------------------------}
  BEGIN
    PLongINT(apData)^ := PRecInfo(apBuffer + FiRecInfoOfs).FiBookMark;
{}END {GetBookmarkData};




{}FUNCTION    TKSS_CustomDataSet.GetBookmarkFlag(apBuffer: PChar): TBookmarkFlag;
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Remarks     - Bookmark flags are used to indicate if a particular record
                is the first or last record in the dataset.  This is necessary
                for "crack" handling.
                If the bookmark flag is bfBOF or bfEOF then the bookmark is
                not actually used; InternalFirst, or InternalLast are called
                instead by TDataSet.
---------------------------------------------------------------------------}
  BEGIN
    result := TBookmarkFlag(0);
    IF apBuffer <> NIL THEN BEGIN
      Result := PRecInfo(apBuffer + FiRecInfoOfs).FeBookMarkFlag;
    END {IF};
{}END {GetBookmarkFlag};




{}FUNCTION  TKSS_CustomDataSet.GetDelimiter : CHAR;
{+O
---------------------------------------------------------------------------
  Purpose     - Return the current delimiter.
---------------------------------------------------------------------------}
  BEGIN
    IF Data <> NIL THEN
      Result := Data.Delimiter
    ELSE BEGIN
      Result := DefaultDelimiter;
    END {IF};
{}END {GetDelimiter};




{}FUNCTION TKSS_CustomDataSet.GetFieldData(aoField: TField; apBuffer: POINTER): BOOLEAN;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Purpose     - copy the data from the record buffer into a field's buffer.
  Remarks     - This function, and SetFieldData, are more complex when
                supporting calculated fields, filters, and other more
                advanced features. See TBDEDataSet for a more complete example.
              - 1999.0108 (KSB) Protected against NIL apBuffer.  
---------------------------------------------------------------------------}
  VAR
    lb        : BOOLEAN;
    lfDateTime: TDateTime;
    lrTimeStamp    : TTimeStamp;
  BEGIN
    IF (apBuffer=NIL) OR (ActiveBuffer=NIL) THEN BEGIN
      result := FALSE;
    END ELSE BEGIN
      CASE aoField.DataType OF
        ftString : BEGIN
          StrLCopy(apBuffer, ActiveBuffer+FaFldOffs^[aoField.FieldNo-1], KSS_GetFieldSize(aoField.FieldNo-1));
          Result := PChar(apBuffer)^ <> #0;
        END {ftString};

        ftDate, ftDateTime,
        ftTime : BEGIN
          Move(PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],lfDateTime,KSS_GetFieldSize(aoField.FieldNo-1));
          lrTimeStamp := DateTimeToTimeStamp(lfDateTime);
          CASE aoField.DataType OF
            ftDate: TDateTimeRec(apBuffer^).Date    := lrTimeStamp.Date;
            ftTime: TDateTimeRec(apBuffer^).Time    := lrTimeStamp.Time;
            ELSE    TDateTimeRec(apBuffer^).DateTime:= TimeStampToMSecs(lrTimeStamp);
          END {CASE};

          Result := TRUE;
        END {ftDate};

        ftBoolean : BEGIN
          Move(PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],lb,SizeOf(lb));
              {the apBuffer^ must be coerced to match the data
               type used in DB's TBooleanField.GetAsString...}
          LongINT(apBuffer^) := LongINT(lb);          {KSB 1998.1221}
//        Bool(apBuffer^) := lb;
          Result := TRUE;
        END {ftBoolean};

        ELSE BEGIN
          Move(PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],PBuffer(apBuffer)^,KSS_GetFieldSize(aoField.FieldNo-1));
          Result := TRUE;
        END {ELSE};
      END {CASE};
    END {IF};
{}END {GetFieldData};




{}PROCEDURE   TKSS_CustomDataSet.GetColumnTypes;
{+O
---------------------------------------------------------------------------
  Purpose     - If the OnGetColumnName event handler has been assigned,
                GetColumnTypes:
                1. passes the name of each column (and handles any changes).
                2. passes an example set of data (up to a maximum of 5 rows)
                3. passes the current data type (default is ftString) (and
                   handles any changes).
  Revised     - 1998.1223 (KSB) Rewrote & Simplified.
              - 1998.1230 (KSB) Added HasHeaderRow logic.
---------------------------------------------------------------------------}
  CONST
    chTAB     = #$09;
  VAR
    liCol     : INTEGER;
    liRow     : INTEGER;
    liRows    : INTEGER;
    leType    : TFieldType;
    lsData    : STRING;
    lsName    : STRING;
  BEGIN
    IF Assigned(FOnGetColName) AND HasHeaderRow THEN BEGIN
              {maximum number of rows from which to extract sample data}
      liRows := Data.RowCount;
      IF liRows > 5 THEN BEGIN
        liRows := 5;
      END {IF};

      FOR liCol := 1 TO Data.ColCount DO BEGIN
              {get the column name}
        lsName := Data.Header.ColumnName[liCol-1];
              {get a data sample}
        lsData := '';
        FOR liRow := 1 TO liRows DO BEGIN
          lsData := lsData + Data.CellAsString[liCol-1,liRow-1];
          IF liRow <> liRows THEN BEGIN
            lsData := lsData + chTab;
          END {IF};
        END {FOR};

              {set the default column data type}
        leType := ftString;
              {let the user change whatever needs changing...}
        FOnGetColName(self,liCol-1,liRows,lsName,lsData,leType);
              {store those changes}
        Data.Header.ColumnName[liCol-1] := lsName;
        Data.Header.ColumnType[liCol-1] := leType;
      END {FOR};
    END {IF};
{}END {GetColumnTypes};




{}FUNCTION    TKSS_CustomDataSet.GetHeader:TColumnDefs;
{+O
---------------------------------------------------------------------------
  Purpose     - Property access method.
---------------------------------------------------------------------------}
  BEGIN
    result := Data.Header;
{}END {GetHeader};




{}FUNCTION    TKSS_CustomDataSet.GetRecNo: LongINT;
{+O
---------------------------------------------------------------------------
  Type        - OPTIONAL.
  Purpose     - Return the current record number.
---------------------------------------------------------------------------}
  BEGIN
    UpdateCursorPos;
    IF (FiCurrentRec = -1) AND (RecordCount > 0) THEN BEGIN
      Result := 1
    END ELSE BEGIN
      Result := FiCurrentRec + 1;
    END {IF};
{}END {GetRecNo};




{}FUNCTION    TKSS_CustomDataSet.GetRecord(apBuffer: PChar; aeGetMode: TGetMode; abDoCheck: BOOLEAN): TGetResult;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Purpose     - 1. It retrieves data for either the current, the prior, or
                   the next record.
                2. It must return the status (TGetResult), and
                3. raise an exception if DoCheck is TRUE.
---------------------------------------------------------------------------}
  BEGIN
    IF RecordCount < 1 THEN
      Result := grEOF
    ELSE BEGIN
      Result := grOK;
      CASE aeGetMode OF
        gmNext:  BEGIN
          IF FiCurrentRec >= RecordCount-1 THEN
            Result := grEOF
          ELSE BEGIN
            Inc(FiCurrentRec);
          END {IF};
        END {gmNext};

        gmPrior: BEGIN
          IF FiCurrentRec <= 0 THEN
            Result := grBOF
          ELSE BEGIN
            Dec(FiCurrentRec);
          END {IF};
        END {gmPrior};

        gmCurrent: BEGIN
          IF (FiCurrentRec < 0) OR (FiCurrentRec >= RecordCount) THEN BEGIN
            Result := grError;
          END {IF};
        END {gmCurrent};
      END {CASE};

              {If the buffers have not been created, don't stuff them with data...}
      IF (apBuffer=NIL) OR (FaFldOffs=NIL) THEN BEGIN {(KSB 1998.1014)}
        Result := grError;
      END ELSE
      IF Result = grOK THEN BEGIN
              {Copy the data into the record buffer...}
        KSS_StoreToBuffer(apBuffer);
      END ELSE
      IF (Result = grError) AND abDoCheck THEN BEGIN
        DatabaseError('No Records');
      END {IF};
    END {IF};
{}END {GetRecord};




{}FUNCTION TKSS_CustomDataSet.GetRecordCount: LongINT;
{+O
---------------------------------------------------------------------------
  Type        - OPTIONAL.
  Purpose     - Return the total number of data records.
---------------------------------------------------------------------------}
  BEGIN
    Result := Data.RowCount;
{}END {GetRecordCount};




{}FUNCTION TKSS_CustomDataSet.GetRecordSize: WORD;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Remarks     - Return just the size of the data in the record buffer.
  Warning     - Do not confuse GetRecordSize with RecBufSize which includes
                additonal structures stored in the record buffer (such as
                TRecInfo).
---------------------------------------------------------------------------}
  BEGIN
    Result := FiRecordSize;
{}END {GetRecordSize};




{}PROCEDURE   TKSS_CustomDataSet.InternalAddRecord(apBuffer: POINTER; abAppend: BOOLEAN);
{+O
---------------------------------------------------------------------------
  Type        - RECORD NAVIGATION/EDITING.
  Remarks     - This method is similar to InternalPost, but the operation
                is always an insert or append and takes a pointer to a
                record buffer as well.
  Revised     - 1999.0211 (KSB) Moved AddRow call to before IF abAppend check.
                Dec'd FiCurrentRec as AddRow is now before InternalLast.
---------------------------------------------------------------------------}
  BEGIN
    FbSaveChanges := TRUE;
    Inc(FiLastBookmark);
    Data.AddRow('', POINTER(FiLastBookmark));
    IF abAppend THEN BEGIN
      InternalLast;
      Dec(FiCurrentRec);
    END {IF};

    IF FiCurrentRec < 0 THEN BEGIN  {1998.1023}
      FiCurrentRec := 0;
    END {IF};

              {store the data from the buffer...}
    KSS_BufferToStore(apBuffer);
{}END {InternalAddRecord};




{}PROCEDURE   TKSS_CustomDataSet.InternalClose;
{+O
---------------------------------------------------------------------------
  Remarks     - Write any edits to disk and free the managing string list.
---------------------------------------------------------------------------}
  BEGIN
    IF NOT FbByPassNIOpen THEN BEGIN
      IF Data.Header.Count > 0 THEN BEGIN
        SaveDataStore;
        IF NOT KeepDataInMemory THEN BEGIN
          Data.Clear;
          Data.Header.Clear;
        END {IF};

              {was: IF DefaultFields THEN...}
        IF FieldCount > 0 THEN BEGIN
          DestroyFields;
        END {IF};

        FreeMem( FaFldOffs );
        FaFldOffs := NIL;
        FiLastBookmark:= 0;
        FiCurrentRec  := -1;
      END {IF};
    END {IF};
{}END {InternalClose};




{}PROCEDURE   TKSS_CustomDataSet.InternalDelete;
{+O
---------------------------------------------------------------------------
  Type        - RECORD NAVIGATION/EDITING.
  Purpose     - Called by TDataSet.Delete to delete the current record
---------------------------------------------------------------------------}
  BEGIN
    FbSaveChanges := TRUE;
    Data.DeleteRow(FiCurrentRec);
    IF FiCurrentRec >= Data.RowCount THEN BEGIN
      Dec(FiCurrentRec);
    END {IF};
{}END {InternalDelete};




{}PROCEDURE   TKSS_CustomDataSet.InternalFirst;
{+O
---------------------------------------------------------------------------
  Type        - RECORD NAVIGATION/EDITING.
  Remarks     - This method is called by TDataSet.First.  Crack behavior is
                required. That is we must position to a special place *before*
                the first record. Otherwise, we will actually end up on the
                second record after Resync is called.
---------------------------------------------------------------------------}
  BEGIN
    FiCurrentRec := -1;
{}END {InternalFirst};




{}PROCEDURE   TKSS_CustomDataSet.InternalGotoBookmark(apBookmark: POINTER);
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Remarks     - In this sample the bookmarks are stored in the Object property
                of the TStringList holding the data.  Positioning to a bookmark
                just requires finding the offset of the bookmark in the
                TStrings.Objects and using that value as the new current record
                pointer.
---------------------------------------------------------------------------}
  VAR
    liIndex   : INTEGER;
    loObject  : TObject;
  BEGIN
    loObject:= TObject(PLongINT(apBookmark)^);
    liIndex := Data.IndexOfObject(loObject);
    IF liIndex <> -1 THEN
      FiCurrentRec := liIndex
    ELSE BEGIN
      DatabaseError('Bookmark not found');
    END {IF};
{}END {InternalGotoBookmark};




{}PROCEDURE   TKSS_CustomDataSet.InternalHandleException;
{+O
---------------------------------------------------------------------------
  Purpose     - This is the exception handler which is called if an exception
                is raised while the component is being stream in or streamed
                out.  In most cases this should be implemented useing the
                application exception handler as follows.
---------------------------------------------------------------------------}
  BEGIN
    Application.HandleException(Self);
{}END {InternalHandleException};




{}PROCEDURE   TKSS_CustomDataSet.InternalInitFieldDefs;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Remarks     - Create one FieldDef for each column found in the data store.
  Revised     - 1998.1023 (KSB) strings are transferred as "PChars" which
                are 1 character longer than the equivalent literal string
                (for the terminating zero). As the length stored in the
                object field of the name has a 5 character pad, we can safely
                reduce the size by 1 byte...
              - 1999.0107 (KSB) Compensated for in KSS_ResetRecordSize.
---------------------------------------------------------------------------}
  CONST
    cbNOTREQUIRED  = FALSE;
  VAR
    liCol     : INTEGER;
    liSize    : INTEGER;
  BEGIN
    FieldDefs.Clear;
    liSize := 0;
    FOR liCol := 1 TO Data.ColCount DO BEGIN
      CASE Data.Header.ColumnType[liCol-1] OF
        ftBoolean,  ftCurrency, ftDate,     ftDateTime,
        ftFloat,    ftInteger,  ftSmallInt, ftTime,
        ftWord   : liSize := 0;
        ftString : liSize := Data.Header.ColumnWidth[liCol-1];//-1;
        ELSE BEGIN
          ColumnCreateError(liCol,Data.Header.ColumnType[liCol-1]);
        END {ELSE};
      END {CASE};

      TFieldDef.Create(FieldDefs, Data.Header.ColumnName[liCol-1], Data.Header.ColumnType[liCol-1], liSize, cbNOTREQUIRED, liCol);
    END {FOR};
{}END {InternalInitFieldDefs};




{}PROCEDURE   TKSS_CustomDataSet.InternalInitRecord(apBuffer: PChar);
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Remarks     - This routine is called to initialize a record buffer.
                We fill the buffer with zeroes, but we might have code to
                initialize default values or do other things as well.
---------------------------------------------------------------------------}
  BEGIN
    FillChar(apBuffer^, RecordSize, 0);
{}END {InternalInitRecord};




{}PROCEDURE   TKSS_CustomDataSet.InternalLast;
{+O
---------------------------------------------------------------------------
  Type        - RECORD NAVIGATION/EDITING.
  Remarks     - Position to the crack *after* the last record.
---------------------------------------------------------------------------}
  BEGIN
    FiCurrentRec := RecordCount;
{}END {InternalLast};




{}PROCEDURE   TKSS_CustomDataSet.InternalOpen;
{+O
---------------------------------------------------------------------------
  Remarks     - This method is called by TDataSet.Open and when FieldDefs
                needs to be updated (usually by the DataSet designer).
                Everything which is allocated or initialized in this method
                should also be freed or uninitialized in the InternalClose
                method.
---------------------------------------------------------------------------}
{}{}PROCEDURE GetFieldDefs;
    CONST
      ciPad   = 5;
    VAR
      liCol        : INTEGER;
      liRow        : INTEGER;
      liMaxWidth   : INTEGER;
    BEGIN
              {Calculate Max Field Width and Record Width}
      FOR liCol := 1 TO Data.ColCount DO BEGIN
              {initialize the maximum width for this column
               with the width of the field (column) name}
        liMaxWidth := Length( Data.Header.ColumnName[liCol-1] );
        FOR liRow := 1 TO Data.RowCount DO BEGIN
          IF Length(Data[liCol-1,liRow-1]) > liMaxWidth THEN BEGIN
            liMaxWidth := Length(Data[liCol-1,liRow-1]);
          END {IF};
        END {FOR};

              {store the maximum field width (plus pad) in
               the header (unless the existing column width
               is not zero)...}
        IF Data.Header.ColumnWidth[liCol-1] = 0 THEN BEGIN      {1999.0107 KSB}
          Data.Header.ColumnWidth[liCol-1] := (liMaxWidth+ciPad);
        END {IF};
      END {FOR};
{}{}END {GetFieldDefs};


  VAR
    li        : INTEGER;
  BEGIN
    IF NOT FbBypassNIOpen THEN BEGIN
              {Load File Into Memory}
      LoadDataStore;
      IF Data.Header.Count > 0 THEN BEGIN
        GetFieldDefs;
        GetColumnTypes;
              {Initialize field definitions and field creation (TFields)}
        InternalInitFieldDefs;
        IF DefaultFields THEN BEGIN
          CreateFields;
        END {IF};

              {force the display widths for the fields...}
        FOR li := 1 TO FieldCount DO BEGIN
          Fields[li-1].DisplayWidth := Data.Header.ColumnWidth[li-1];
        END {FOR};

        BindFields(TRUE);
        KSS_ResetRecordSize;

              {create Bookmarks}
        FOR li := 1 TO Data.RowCount DO BEGIN
          Data.RowObjects[li-1] := POINTER(li);
        END {FOR};

        FiCurrentRec  := -1;
        FiLastBookmark:= Data.RowCount;
        BookmarkSize  := SizeOf( LongINT );

        {$IFDEF Debug}
        ReportTo(Name+'.txt',FALSE);
        {$ENDIF}
      END {IF};
    END {IF};
{}END {InternalOpen};




{}PROCEDURE   TKSS_CustomDataSet.InternalPost;
{+O
---------------------------------------------------------------------------
  Type        - RECORD NAVIGATION/EDITING.
  Purpose     - Called by TDataSet.Post.
  Remarks     - Most implmentations would write the changes directly to the
                associated datasource, but here we simply set a flag to write
                the changes when we close the dateset.
  Revised     - 1999.0108 (KSB) Moved the "IF FiCurrentRec < 0" outside of the
                "IF State<> dsEdit" block.
              - 1999.0211 (KSB) Added code to reset FiCurrentRec to the "end of file".
---------------------------------------------------------------------------}
  BEGIN
    FbSaveChanges := TRUE;
    IF State<>dsEdit THEN BEGIN
      Inc(FiLastBookmark);
      Data.AddRow('', POINTER(FiLastBookmark));
    END {IF};

    IF FiCurrentRec < 0 THEN BEGIN  {1998.1023}
      FiCurrentRec := 0;
    END ELSE BEGIN                  {1999.0211}
      FiCurrentRec := Data.RowCount-1;
    END {IF};

    KSS_BufferToStore(ActiveBuffer);
{}END {InternalPost};




{}PROCEDURE   TKSS_CustomDataSet.InternalSetToRecord(apBuffer: PChar);
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Purpose     - This function does the same thing as InternalGotoBookmark,
                but it takes a record buffer as a parameter instead
---------------------------------------------------------------------------}
  BEGIN       {guard against dereferencing a nil buffer... (KSB)}
    IF apBuffer <> NIL THEN BEGIN
      InternalGotoBookmark(@PRecInfo(apBuffer + FiRecInfoOfs).FiBookMark);
    END {IF};
{}END {InternalSetToRecord};




{}FUNCTION TKSS_CustomDataSet.IsCursorOpen: BOOLEAN;
{+O
---------------------------------------------------------------------------
  Remarks     - Used while opening the dataset.
                It indicates if data is available even though the
                current state is still dsInActive.
---------------------------------------------------------------------------}
  BEGIN
    Result := Assigned(Data) AND (FieldCount>0);
{}END {IsCursorOpen};




{}PROCEDURE   TKSS_CustomDataSet.KSS_BufferToStore(apBuffer:PChar);
{+O
---------------------------------------------------------------------------
  Purpose     - Move data from the record buffer to the data store (FoData).
  Remarks     - Non-string data is converted to hexadecimal strings for
                storage.
---------------------------------------------------------------------------}
  VAR
    lb        : BOOLEAN;
    lf        : DOUBLE;
    lfDollar  : CURRENCY;
    li        : INTEGER;
    liInt     : LongINT;
    liSmall   : SmallINT;
    liWord    : WORD;
  BEGIN
    FOR li := 1 TO Data.ColCount DO BEGIN
      CASE Fields[li-1].DataType OF
        ftString : Data[li-1,FiCurrentRec] := StrPas(PChar(apBuffer)+FaFldOffs^[li-1]);

        ftInteger: BEGIN
          Move(PBuffer(apBuffer)^[FaFldOffs^[li-1]],liInt,KSS_GetFieldSize(li-1));
          Data.IntegerCell[li-1,FiCurrentRec] := liInt;
        END {ftInteger};

        ftFloat: BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],lf,KSS_GetFieldSize(li-1));
          Data.FloatCell[li-1,FiCurrentRec] := lf;
        END {ftFloat};

        ftDate,ftDateTime,
        ftTime : BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],lf,KSS_GetFieldSize(li-1));
          CASE Fields[li-1].DataType OF
            ftDate : Data.DateCell[li-1,FiCurrentRec] := lf;
            ftTime : Data.TimeCell[li-1,FiCurrentRec] := lf;
            ELSE     Data.DateTimeCell[li-1,FiCurrentRec] := lf;
          END {CASE};
        END {ftDate,ftDateTime,ftTime};

        ftBoolean : BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],lb,SizeOf(lb));
          Data.BooleanCell[li-1,FiCurrentRec] := lb;
        END {ftBoolean};

        ftSmallInt : BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],liSmall,KSS_GetFieldSize(li-1));
          Data.SmallIntCell[li-1,FiCurrentRec] := liSmall;
        END {ftShortInt};

        ftWord : BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],liWord,KSS_GetFieldSize(li-1));
          Data.WordCell[li-1,FiCurrentRec] := liWord;
        END {ftWord};

        ftCurrency : BEGIN
          Move(PBuffer(apBuffer)^[faFldOffs^[li-1]],lf,KSS_GetFieldSize(li-1));
          lfDollar := lf;
          Data.CurrencyCell[li-1,FiCurrentRec] := lfDollar;
        END {ftCurrency};
      END {CASE};
    END {FOR};
{}END {KSS_BufferToStore};




{}FUNCTION    TKSS_CustomDataSet.KSS_GetFieldSize(aiFieldNo:INTEGER):INTEGER;
{+O
---------------------------------------------------------------------------
  Purpose     - Return the size in bytes of the data stored in the record buffer.
  Revised     - 1998.1228 (KSB) Added code to return column width if fields have been destroyed. 
---------------------------------------------------------------------------}
  BEGIN
    result := 0;
    CASE Data.Header.ColumnType[aiFieldNo] OF
      ftInteger    : result := SizeOf(INTEGER);
      ftDate,ftDateTime,ftFloat,
      ftTime       : result := SizeOf(DOUBLE);
      ftBoolean    : result := SizeOf(BOOLEAN);
      ftSmallInt   : result := SizeOf(SmallINT);
      ftWord       : result := SizeOf(WORD);
      ftCurrency   : result := SizeOf(CURRENCY);
      ftString     : BEGIN
        IF FieldCount = 0 THEN BEGIN
          result := Data.Header.ColumnWidth[aiFieldNo];
        END ELSE BEGIN
          result := Fields[aiFieldNo].Size;
        END {IF};
      END {ftString};
    END {CASE};
{}END {KSS_GetFieldSize};




{}PROCEDURE   TKSS_CustomDataSet.KSS_ResetRecordSize;
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
  Purpose     - Compute the size of the record needed to contain all the
                binary raw data for a row. Adjust the offsets to point to
                the proper start of each field.
  Revised     - 1998.1023 (KSB) strings are transferred as "PChars" which
                are 1 character longer than the equivalent literal string
                (for the terminating zero). As the length stored returned
                by the KSS_GetFieldSize decrements the string size by 1,
                we have to increase the offset by 1 so the zero byte of
                the current string does not overwrite the first character
                of the next field...
              - 1999.0107 (KSB) Rewrote to compute both the values for
                FaFldOffs and FiRecInfoOfs.
---------------------------------------------------------------------------}
  VAR
    liCol     : INTEGER;
    liSize    : INTEGER;
  BEGIN
    IF Assigned( FaFldOffs ) THEN BEGIN
      FreeMem( FaFldOffs );
    END {IF};

              {Create array of offsets for internal use...
               Each [i] element is the offset of field i as
               measured from the start of the record buffer}
    GetMem( FaFldOffs, SizeOf(INTEGER)*Data.ColCount );
    FiRecordSize := 0;
    FOR liCol := 1 TO Data.ColCount DO BEGIN
      FaFldOffs^[liCol-1] := FiRecordSize;
      liSize := KSS_GetFieldSize(liCol-1);
      IF Fields[liCol-1].DataType = ftString THEN BEGIN
        Inc(liSize);
      END {IF};

              {increment total record size by the width of the current column}
      Inc(FiRecordSize, liSize);
    END {FOR};

    FiRecInfoOfs := FiRecordSize;
    FiRecBufSize := FiRecInfoOfs + SizeOf(TRecInfo);
{}END {KSS_ResetRecordSize};




{}PROCEDURE   TKSS_CustomDataSet.KSS_StoreToBuffer(apBuffer:PChar);
{+O
---------------------------------------------------------------------------
  Purpose     - Move data from the store (FoData) to the buffer.
  Remarks     - Non-string data is converted from hexadecimal strings to
                the appropriate binary value before placed in the buffer.
---------------------------------------------------------------------------}
  VAR
    lb        : BOOLEAN;
    lf        : DOUBLE;
    lfDollar  : CURRENCY;
    li        : INTEGER;
    liInt     : LongINT;
    liSmall   : SmallINT;
    liWord    : WORD;
  BEGIN
    FOR li := 1 TO Data.ColCount DO BEGIN
      CASE Fields[li-1].DataType OF
        ftString : BEGIN
          StrLCopy(apBuffer+FaFldOffs^[li-1], PChar(Data[li-1,FiCurrentRec]), KSS_GetFieldSize(li-1));
        END {ftString};

        ftInteger: BEGIN
          liInt := Data.IntegerCell[li-1,FiCurrentRec];
          Move(liInt,PBuffer(apBuffer)^[FaFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftInteger};

        ftFloat : BEGIN
          lf := Data.FloatCell[li-1,FiCurrentRec];
          Move(lf,PBuffer(apBuffer)^[faFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftFloat};

        ftDate,ftDateTime,
        ftTime : BEGIN
          lf := Data.DateTimeCell[li-1,FiCurrentRec];
          Move(lf,PBuffer(apBuffer)^[faFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftDate,ftDateTime,ftTime};

        ftBoolean : BEGIN
          lb := Data.BooleanCell[li-1,FiCurrentRec];
          Move(lb,PBuffer(apBuffer)^[faFldOffs^[li-1]],SizeOf(lb));
        END {ftBoolean};

        ftSmallInt : BEGIN
          liSmall := Data.SmallIntCell[li-1,FiCurrentRec];
          Move(liSmall,PBuffer(apBuffer)^[faFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftShortInt};

        ftWord : BEGIN
          liWord := Data.WordCell[li-1,FiCurrentRec];
          Move(liWord,PBuffer(apBuffer)^[faFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftWord};

        ftCurrency : BEGIN
          lfDollar := Data.CurrencyCell[li-1,FiCurrentRec];
          lf := lfDollar;
          Move(lf,PBuffer(apBuffer)^[faFldOffs^[li-1]],KSS_GetFieldSize(li-1));
        END {ftCurrency};
      END {CASE};
    END {FOR};

    WITH PRecInfo(apBuffer + FiRecInfoOfs)^ DO BEGIN
      FeBookMarkFlag := bfCurrent;
      FiBookMark := LongINT(Data.RowObjects[FiCurrentRec]);
    END {WITH};
{}END {KSS_StoreToBuffer};




{}PROCEDURE   TKSS_CustomDataSet.LoadDataStore;
{+O
---------------------------------------------------------------------------
  Type        - VIRTUAL.
  Purpose     - Write code in a descendent to load data to the data store.
---------------------------------------------------------------------------}
  BEGIN
{}END {LoadDataSet};




{}PROCEDURE   TKSS_CustomDataSet.ReadFld(aoReader: TReader);
{+O
---------------------------------------------------------------------------
  Purpose     - Read the header property.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    aoReader.ReadListBegin;
    Data.Header.BeginUpdate;
    TRY
      Data.Header.Clear;
      li := 0;
      WHILE NOT aoReader.EndOfList DO BEGIN
        Data.Header.AsString[li] := aoReader.ReadString;
        Inc(li);
      END {WHILE};
    FINALLY
      Data.Header.EndUpdate;
    END {TRY};

    aoReader.ReadListEnd;
{}END {ReadFld};




{}PROCEDURE   TKSS_CustomDataSet.Rebookmark;
{+O
---------------------------------------------------------------------------
  Purpose     - Recreate the internal bookmarks.
  Remarks     - Should be called after loading data directly into the store.
  Revised     - 1998.1117 (KSB) Wrote initial version.
---------------------------------------------------------------------------}
  BEGIN
    FiCurrentRec  := -1;
    FiLastBookMark:= 0;
    WHILE FiLastBookMark < data.RowCount DO BEGIN
      data.RowObjects[FiLastBookMark] := POINTER(FiLastBookMark);
      Inc(FiLastBookMark);
    END {WHILE};
{}END {Rebookmark};




{}PROCEDURE   TKSS_CustomDataSet.RefreshDisplayLabels;
{+O
---------------------------------------------------------------------------
  Purpose     - Refresh the display labels. Generally done if the column
                name has been changed by the user.
---------------------------------------------------------------------------}
  VAR
    liCol     : INTEGER;
  BEGIN
    DestroyFields;
    FbKeepDataInMem := TRUE;
    FbBypassNIOpen := TRUE;
    active := FALSE;
    InternalInitFieldDefs;
    CreateFields;
              {force the display widths for the fields...}
    FOR liCol := 1 TO FieldCount DO BEGIN
      Fields[liCol-1].DisplayWidth := Data.Header.ColumnWidth[liCol-1];
    END {FOR};

    BindFields(TRUE);
    KSS_ResetRecordSize;
    active := TRUE;
    FbBypassNIOpen := FALSE;
    FbSaveChanges := TRUE;
{}END {RefreshDisplayLabels};




{}PROCEDURE   TKSS_CustomDataSet.Report;
{+O
---------------------------------------------------------------------------
  Purpose     - Write the current data set status to its report file.
---------------------------------------------------------------------------}
  BEGIN
    ReportTo(name+'.txt',TRUE);
{}END {Report};




{}PROCEDURE   TKSS_CustomDataSet.ReportTo(asFileName:STRING; abAppend:BOOLEAN);
{+O
---------------------------------------------------------------------------
  Purpose     - Dump the header & data for inspection.
  Revised     - 1998.1106 (KSB) Fixed formatting.
              - 1999.0426 (KSB) Added data set name to report header.
---------------------------------------------------------------------------}
  VAR
    liCol     : INTEGER;
    liNext    : INTEGER;
    liRow     : INTEGER;
    liWidth   : INTEGER;
    loEnums   : TKustomEnum;
    ls        : STRING;
    lTXT      : TEXT;
  BEGIN
    TRY
      AssignFile(lTXT,asFileName);
      IF abAppend THEN BEGIN
        IF FileExists(asFileName) THEN BEGIN
          System.Append(lTXT);
        END ELSE BEGIN
          System.Rewrite(lTXT);
        END {IF};
      END ELSE BEGIN
        Rewrite(lTXT);
      END {IF};

      WriteLn(lTXT,'        KSS DATA SET REPORT (',Name,')');
      WriteLn(lTXT,'        ======================',StringOfChar('=',Length(Name)));
      WriteLn(lTXT,'          ',FormatDateTime('yyyymmdd.hhnnss',Now));
      WriteLn(lTXT,'       Has Header Row = ',HasHeaderRow);
      WriteLn(lTXT,'              Columns');
      WriteLn(lTXT,'Offset  Size  Next   Max  Type        Name    ');
                   {  xxxx   xxx  xxxx  xxxx  ftCurrency  name}
      loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
      TRY
        FOR liCol := 1 TO Data.ColCount DO BEGIN
          ls := loEnums.ToString(Ord(Data.Header.ColumnType[liCol-1]));
          liNext := FaFldOffs[liCol-1] + KSS_GetFieldSize(liCol-1);
          Write(lTXT,FaFldOffs[liCol-1]:6);
          Write(lTXT,KSS_GetFieldSize(liCol-1):6);
          Write(lTXT,liNext:6);
          Write(lTXT,Data.Header.ColumnWidth[liCol-1]:6);
          Write(lTXT,'  ',ls:10);
          Write(lTXT,'  ',Data.Header.ColumnName[liCol-1]);
          WriteLn(lTXT);
        END {FOR};
      FINALLY
        loEnums.Free;
        Flush(lTXT);
      END {TRY};

      WriteLn(lTXT);
      WriteLn(lTXT,'Row    KSS DATA STORE REPORT');
      WriteLn(lTXT,'===    =====================');
      FOR liRow := 1 TO Data.RowCount DO BEGIN
        Write(lTXT,liRow:3,'    ');
        FOR liCol := 1 TO Data.ColCount DO BEGIN
          liWidth := KSS_GetFieldSize(liCol-1);
          IF Data.Header.ColumnType[liCol-1] <> ftString THEN BEGIN
            liWidth := Length(TStorage(Data).Cell[liCol-1,liRow-1]);
          END {IF};

          liWidth := liWidth+2;
          Write(lTXT,TStorage(Data).Cell[liCol-1,liRow-1]:liWidth);
        END {FOR};

        WriteLn(lTXT);
        Write(lTXT,'       ');
        FOR liCol := 1 TO Data.ColCount DO BEGIN
          liWidth := KSS_GetFieldSize(liCol-1);
          IF Data.Header.ColumnType[liCol-1] <> ftString THEN BEGIN
            liWidth := Length(TStorage(Data).Cell[liCol-1,liRow-1]);
          END {IF};

          liWidth := liWidth+2;
          Write(lTXT,TStorage(Data).CellAsString[liCol-1,liRow-1]:liWidth);
        END {FOR};

        WriteLn(lTXT);
      END {FOR};
    FINALLY
      CloseFile(lTXT);
    END {TRY};
{}END {ReportTo};




{}PROCEDURE   TKSS_CustomDataSet.SaveDataStore;
{+O
---------------------------------------------------------------------------
  Type        - VIRTUAL.
  Purpose     - Write code in a descendent to save data to a file.
---------------------------------------------------------------------------}
  BEGIN
{}END {SaveDataStore};




{}PROCEDURE   TKSS_CustomDataSet.SetBookmarkData(apBuffer: PChar; apData: POINTER);
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Purpose     - Provides a way to write bookmark data into the record buffer
                without actually repositioning the current record
---------------------------------------------------------------------------}
  BEGIN
    PRecInfo(apBuffer + FiRecInfoOfs).FiBookMark := PLongINT(apData)^;
{}END {SetBookmarkData};




{}PROCEDURE   TKSS_CustomDataSet.SetBookmarkFlag(apBuffer: PChar; aeValue: TBookmarkFlag);
{+O
---------------------------------------------------------------------------
  Type        - BOOKMARK.
  Purpose     - Set the bookmark flag in the record buffer.
---------------------------------------------------------------------------}
  BEGIN
    PRecInfo(apBuffer + FiRecInfoOfs).FeBookMarkFlag := aeValue;
{}END {SetBookmarkFlag};




{}PROCEDURE   TKSS_CustomDataSet.SetDelimiter( acDelimiter : CHAR );
{+O
---------------------------------------------------------------------------
  Purpose     - Set the data delimiter.
---------------------------------------------------------------------------}
  BEGIN
    IF Data <> NIL THEN BEGIN
      Data.Delimiter := acDelimiter;
    END {IF};
{}END {SetDelimiter};




{}PROCEDURE   TKSS_CustomDataSet.SetFieldData(aoField: TField; apBuffer: POINTER);
{+O
---------------------------------------------------------------------------
  Type        - RECORD/FIELD ACCESS.
---------------------------------------------------------------------------}
  VAR
    lb        : BOOLEAN;
    lfDateTime: TDateTime;
    lrTimeStamp    : TTimeStamp;
  BEGIN
    CASE aoField.DataType OF
      ftString : BEGIN
        StrLCopy(ActiveBuffer+FaFldOffs^[aoField.FieldNo-1], apBuffer, KSS_GetFieldSize(aoField.FieldNo-1));
      END {ftString};

      ftBoolean : BEGIN
              {the apBuffer^ must be coerced to match the data
               type used in DB's TBooleanField.SetAsBoolean...}
        lb := WordBool(apBuffer^);
        Move(lb,PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],SizeOf(lb));
      END {ftBoolean};

      ftDate, ftDateTime,
      ftTime : BEGIN
        CASE aoField.DataType OF
          ftDate: BEGIN
            lrTimeStamp.Time := 0;
            lrTimeStamp.Date := TDateTimeRec(apBuffer^).Date;
          END {ftDate};

          ftTime: BEGIN
            lrTimeStamp.Time := TDateTimeRec(apBuffer^).Time;
            lrTimeStamp.Date := DateDelta;
          END {ftTime};

          ELSE BEGIN
            TRY
              lrTimeStamp := MSecsToTimeStamp(TDateTimeRec(apBuffer^).DateTime);
            EXCEPT
              lrTimeStamp.Time := 0;
              lrTimeStamp.Date := 0;
            END {TRY};
          END {ELSE};
        END {CASE};

        lfDateTime := TimeStampToDateTime(lrTimeStamp);
        Move(lfDateTime,PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],SizeOf(lfDateTime));
      END {ftDate,ftDateTime,ftTime};

      ELSE BEGIN
        Move(PBuffer(apBuffer)^,PBuffer(ActiveBuffer)^[FaFldOffs^[aoField.FieldNo-1]],KSS_GetFieldSize(aoField.FieldNo-1));
      END {ELSE};
    END {CASE};

    DataEvent(deFieldChange, LongINT(aoField));
{}END {SetFieldData};




{}PROCEDURE   TKSS_CustomDataSet.SetHeader(aoHeader:TColumnDefs);
{+O
---------------------------------------------------------------------------
  Purpose     - Property access method.
---------------------------------------------------------------------------}
  BEGIN
    Data.Header := aoHeader;
{}END {SetHeader};




{}PROCEDURE   TKSS_CustomDataSet.SetRecNo(aiValue: INTEGER);
{+O
---------------------------------------------------------------------------
  Type        - OPTIONAL.
  Purpose     - Set the current record number.
---------------------------------------------------------------------------}
  BEGIN
    IF (aiValue >= 0) AND (aiValue < Data.RowCount-1) THEN BEGIN
      FiCurrentRec := aiValue - 1;
      Resync([]);
    END {IF};
{}END {SetRecNo};




{}PROCEDURE   TKSS_CustomDataSet.WriteFld( aoWriter: TWriter );
{+O
---------------------------------------------------------------------------
  Purpose     - Store the header property.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    aoWriter.WriteListBegin;
    FOR li := 1 TO Data.Header.Count DO BEGIN
//    aoWriter.WriteString( Data.Header.ColumnName[li-1] );
      aoWriter.WriteString( Data.Header.AsString[li-1] );
    END {FOR};

    aoWriter.WriteListEnd;
{}END {WriteFld};


              {TextTable Methods}


{}PROCEDURE   TTextTable.LoadDataStore;
{+O
---------------------------------------------------------------------------
  Purpose     - Load data from the file to the data store.
  Revised     - 1998.1230 (KSB) Added HasHeaderRow logic.
---------------------------------------------------------------------------}
  VAR
    liCol     : INTEGER;
    loList    : TList;
    lpSave    : TTypeChangeEvent;
  BEGIN
    TRY
      Data.AllowBlankLines := FbBlankLinesOk;
      Data.LoadFromFile(FileName);
      IF HasHeaderRow THEN BEGIN
        Data.Header.Clear;
              {row zero is assumed to contain only the column names...
               so copy them to the field names array property...}
        FOR liCol := 1 TO Data.ColCount DO BEGIN
          Data.Header.Add( Data[liCol-1,0] );
        END {FOR};
              {...and remove them from the data...}
        Data.DeleteRow(0);
      END ELSE BEGIN
        lpSave := Data.Header.OnTypeChange;
        Data.Header.OnTypeChange := NIL;

        IF Data.ColCount > Data.Header.Count THEN BEGIN
          FOR liCol := Data.Header.Count TO Data.ColCount DO BEGIN
            Data.Header.AsString[liCol] := 'Col.'+IntToStr(liCol)+',ftString,20';
          END {FOR};
        END {IF};

        loList := TList.Create;
        TRY
          FOR liCol := 1 TO Data.Header.Count DO BEGIN
            loList.Add(POINTER(Data.Header.ColumnType[liCol-1]));
            Data.Header.ColumnType[liCol-1] := ftString;
          END {FOR};

          Data.Header.OnTypeChange := lpSave;
          FOR liCol := 1 TO Data.Header.Count DO BEGIN
            IF TFieldType(loList[liCol-1]) <> ftString THEN BEGIN
              Data.Header.ColumnType[liCol-1] := TFieldType(loList[liCol-1]);
            END {IF};
          END {FOR};
        FINALLY
          loList.Free;
          Data.Header.OnTypeChange := lpSave;
        END {IF};
      END {IF};
    EXCEPT
    END {TRY};
{}END {LoadDataStore};




{}PROCEDURE   TTextTable.SaveDataStore;
{+O
---------------------------------------------------------------------------
  Purpose     - Write changes back to the text file.
  Revised     - 1998.1230 (KSB) Added HasHeaderRow logic.
---------------------------------------------------------------------------}
  VAR
    liCol     : INTEGER;
  BEGIN
    TRY
      IF (FileName <> '') AND SaveChanges THEN BEGIN
              {convert back to string representation}
        FOR liCol := 1 TO Data.Header.Count DO BEGIN
          IF Data.Header.ColumnType[liCol-1] <> ftString THEN BEGIN
            Data.Header.ColumnType[liCol-1] := ftString;
          END {IF};
        END {FOR};

              {add column names -- iff it has a header row}
        IF HasHeaderRow THEN BEGIN
          Data.AddRow('', POINTER(0) );
          FOR liCol :=1 TO Data.Header.Count DO BEGIN
            Data[liCol-1,0] := Data.Header.ColumnName[liCol-1];
          END {FOR};
        END {IF};

              {write the data to the file}
        Data.SaveToFile(FileName);

              {remove the column names}
        IF HasHeaderRow THEN BEGIN
          Data.DeleteRow( 0 );
        END {IF};
      END {IF};
    EXCEPT
    END {TRY};
{}END {SaveDataStore};




END {TxtTable}.
