UNIT KSSDates;
{+H
---------------------------------------------------------------------------
  File        - KSSDATES.PAS
  Purpose     - Implements modified String to Date/Time functions
  Reference   - Delphi's SysUtils.PAS code.
  Copyright    1998 Keith S. Brown, except as noted above.
  Revised     - 1998.1023(KSB) Wrote original version.
---------------------------------------------------------------------------}
INTERFACE

{}FUNCTION    KSS_StrToDate(CONST asStr: STRING): TDateTime;
{}FUNCTION    KSS_StrToDateTime(CONST asStr:STRING):TDateTime;
{}FUNCTION    KSS_StrToTime(CONST asStr: STRING): TDateTime;
              {=================================================}
IMPLEMENTATION
USES
  Windows,  SysUtils;


{}PROCEDURE   ConvertErrorFmt(CONST asIdent:STRING; CONST aArgs:ARRAY OF CONST);
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  BEGIN
    RAISE EConvertError.CreateFmt(asIdent, aArgs);
{}END {ConvertErrorFmt};




{}FUNCTION    CurrentYear: WORD;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    lrSystemTime   : TSystemTime;
  BEGIN
    GetLocalTime(lrSystemTime);
    Result := lrSystemTime.wYear;
{}END {CurrentYear};




{}FUNCTION    DoEncodeDate(aiYear, aiMonth, aiDay:WORD; VAR afDate:TDateTime): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    li        : Integer;
    lpDayTable: PDayTable;
  BEGIN
    Result := FALSE;
    lpDayTable := @MonthDays[IsLeapYear(aiYear)];
    IF (aiYear >= 1) AND (aiYear <= 9999) AND (aiMonth >= 1) AND (aiMonth <= 12)
    AND (aiDay >= 1) AND (aiDay <= lpDayTable^[aiMonth]) THEN BEGIN
      FOR li := 1 TO aiMonth - 1 DO BEGIN
        Inc(aiDay, lpDayTable^[li]);
      END {FOR};

      li := aiYear - 1;
      afDate := li * 365 + li div 4 - li div 100 + li div 400 + aiDay - DateDelta;
      Result := TRUE;
    END {IF};
{}END {DoEncodeDate};




{}FUNCTION    DoEncodeTime(aiHour,aiMinute,aiSecond,aiMSec:Word; VAR afTime:TDateTime):BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  BEGIN
    Result := FALSE;
    IF (aiHour < 24) AND (aiMinute < 60) AND (aiSecond < 60) AND (aiMSec < 1000) THEN BEGIN
      afTime := (aiHour * 3600000 + aiMinute * 60000 + aiSecond * 1000 + aiMSec) / MSecsPerDay;
      Result := TRUE;
    END {IF};
{}END {DoEncodeTime};


{ String to date/time conversions }


TYPE
  TDateOrder = (doMDY, doDMY, doYMD);


{}FUNCTION    GetDateOrder(CONST DateFormat: STRING): TDateOrder;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    Result := doMDY;
    li := 1;
    WHILE li <= Length(DateFormat) DO BEGIN
      CASE Chr(Ord(DateFormat[li]) AND $DF) OF
        'E': Result := doYMD;
        'Y': Result := doYMD;
        'M': Result := doMDY;
        'D': Result := doDMY;
        ELSE BEGIN
          Inc(li);
          Continue;
        END {ELSE};
      END {CASE};

      Exit;
    END {WHILE};

    Result := doMDY;
{}END {GetDateOrder};




{}FUNCTION    GetPattern(asFormat:STRING):STRING;
{+H
---------------------------------------------------------------------------
  Remarks     - Return a string representing the pattern in the format
                string "asFormat".
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    result := '';
    FOR li := 1 TO Length(asFormat) DO BEGIN
      CASE UpCase(asFormat[li]) OF
        '0'..'9' : result := result + 'N';
        'A'..'Z' : result := result + 'T';
        ':',' '  : result := result + asFormat[li];
        '/','-'  : result := result + '-';
        ELSE       result := result + asFormat[li];
      END {CASE};
    END {FOR};
{}END {GetPattern};




{}FUNCTION    MonthNo(asStr:STRING):INTEGER;
{+H
---------------------------------------------------------------------------
  Remarks     - Return the month number for a 3 character month abbrev.
---------------------------------------------------------------------------}
  CONST                  {....^....1....^....2....^....3....^.}
    csMonths  : STRING = 'JANFEBMARAPRMAYJUNJULAUGSEPOCTNOVDEC';
  BEGIN                  {.  .  .  .  .  .  .  .  .  .  .  .  }
    result := 0;
    IF Length(asStr) >= 3 THEN BEGIN
      asStr := Copy(UpperCase(asStr),1,3);
      result := (Pos(asStr,csMonths)+2) DIV 3;
    END {IF};
{}END {MonthNo};




{}PROCEDURE   ScanBlanks(CONST asStr: STRING; VAR aiPos: INTEGER);
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    li := aiPos;
    WHILE (li <= Length(asStr)) AND (asStr[li] = ' ') DO BEGIN
      Inc(li);
    END {WHILE};

    aiPos := li;
{}END {ScanBlanks};




{}FUNCTION    ScanChar(CONST asStr: STRING; VAR aiPos: INTEGER; acCh: Char): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  BEGIN
    Result := FALSE;
    ScanBlanks(asStr, aiPos);
    IF (aiPos <= Length(asStr)) AND (asStr[aiPos] = acCh) THEN BEGIN
      Inc(aiPos);
      Result := TRUE;
    END {IF};
{}END {ScanChar};




{}FUNCTION    ScanNumber(CONST asStr: STRING; VAR aiPos: INTEGER; VAR aiNumber: WORD): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
    liN       : WORD;
  BEGIN
    Result := FALSE;
    ScanBlanks(asStr, aiPos);
    li  := aiPos;
    liN := 0;
    WHILE (li <= Length(asStr)) AND (asStr[li] in ['0'..'9']) AND (liN < 1000) DO BEGIN
      liN := liN * 10 + (Ord(asStr[li]) - Ord('0'));
      Inc(li);
    END {WHILE};

    IF li > aiPos THEN BEGIN
      aiPos   := li;
      aiNumber:= liN;
      Result  := TRUE;
    END {IF};
{}END {ScanNumber};




{}FUNCTION    ScanString(CONST asStr: STRING; VAR aiPos: INTEGER; CONST Symbol: STRING): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  BEGIN
    Result := FALSE;
    IF Symbol <> '' THEN BEGIN
      ScanBlanks(asStr, aiPos);
      IF AnsiCompareText(Symbol, Copy(asStr, aiPos, Length(Symbol))) = 0 THEN BEGIN
        Inc(aiPos, Length(Symbol));
        Result := TRUE;
      END {IF};
    END {IF};
{}END {ScanString};




{}PROCEDURE   ScanToNumber(CONST asStr: STRING; VAR aiPos: INTEGER);
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  BEGIN
    WHILE (aiPos <= Length(asStr)) AND NOT (asStr[aiPos] IN ['0'..'9']) DO BEGIN
      IF asStr[aiPos] IN LeadBytes THEN BEGIN
        Inc(aiPos);
      END {IF};

      Inc(aiPos);
    END {WHILE};
{}END {ScanToNumber};




{}FUNCTION    TranslateFormat(asDateTime:STRING; VAR aeDateOrder:TDateOrder):STRING;
{+H
---------------------------------------------------------------------------
  Remarks     - Return the most probable date order for a pattern that
                contains a 3 character month abbreviation.
                The return value is a date string. If the asDateTime parameter
                contains a 3 character month abbrev., the returned string
                translates that to a dd mm yyyy format.
---------------------------------------------------------------------------}
  VAR
    ls        : STRING;
  BEGIN
    ls := GetPattern(asDateTime);
    IF (Copy(ls,1,11) = 'NN-TTT-NNNN') THEN BEGIN
      result := Copy(asDateTime,1,3)+IntToStr(MonthNo(Copy(asDateTime,4,3)))+Copy(asDateTime,7,Length(asDateTime));
      aeDateOrder := doDMY;
    END ELSE
    IF (Copy(ls,1,10) = 'N-TTT-NNNN') THEN BEGIN
      result := '0'+Copy(asDateTime,1,2)+IntToStr(MonthNo(Copy(asDateTime,3,3)))+Copy(asDateTime,6,Length(asDateTime));
      aeDateOrder := doDMY;
    END ELSE
    IF (Copy(ls,1,09) = 'NN-TTT-NN') THEN BEGIN
      result := Copy(asDateTime,1,3)+IntToStr(MonthNo(Copy(asDateTime,4,3)))+Copy(asDateTime,7,Length(asDateTime));
      aeDateOrder := doDMY;
    END ELSE
    IF (Copy(ls,1,08) = 'N-TTT-NN') THEN BEGIN
      result := '0'+Copy(asDateTime,1,2)+IntToStr(MonthNo(Copy(asDateTime,3,3)))+Copy(asDateTime,6,Length(asDateTime));
      aeDateOrder := doDMY;
    END ELSE BEGIN
      result := asDateTime;
      aeDateOrder := GetDateOrder(ShortDateFormat);
    END {IF};
{}END {TranslateFormat};




{}FUNCTION    ScanDate(CONST asStr:STRING; aeDateOrder:TDateOrder; VAR aiPos:INTEGER; VAR afDate:TDateTime): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Modified from the ScanDate found in SysUtils to accept a
                value of date order computed outside the function.
---------------------------------------------------------------------------}
  VAR
    liDay          : WORD;
    liMonth        : WORD;
    liN1           : WORD;
    liN2           : WORD;
    liN3           : WORD;
    liYear         : WORD;
  BEGIN
    liYear := 0;
    liMonth:= 0;
    liDay  := 0;
    Result := FALSE;
              {skip over prefix text}
    IF ShortDateFormat[1] = 'g' THEN BEGIN
      ScanToNumber(asStr, aiPos);
    END {IF};

    IF NOT(ScanNumber(asStr,aiPos,liN1)
       AND ScanChar  (asStr,aiPos,DateSeparator)
       AND ScanNumber(asStr,aiPos,liN2))THEN BEGIN
      Exit;
    END {IF};

    IF ScanChar(asStr, aiPos, DateSeparator) THEN BEGIN
      IF NOT ScanNumber(asStr, aiPos, liN3) THEN BEGIN
        Exit;
      END {IF};

      CASE aeDateOrder OF
        doMDY: BEGIN
          liYear := liN3;
          liMonth:= liN1;
          liDay  := liN2;
        END {doMDY};

        doDMY: BEGIN
          liYear := liN3;
          liMonth:= liN2;
          liDay  := liN1;
        END {doDMY};

        doYMD: BEGIN
          liYear := liN1;
          liMonth:= liN2;
          liDay  := liN3;
        END {doYMD};
      END {CASE};

      IF liYear <= 99 THEN BEGIN
        Inc(liYear, CurrentYear DIV 100 * 100);
      END {IF};
    END ELSE BEGIN
      liYear := CurrentYear;
      IF aeDateOrder = doDMY THEN BEGIN
        liDay  := liN1;
        liMonth:= liN2;
      END ELSE BEGIN
        liMonth:= liN1;
        liDay  := liN2;
      END {IF};
    END {IF};

    ScanChar(asStr, aiPos, DateSeparator);
    ScanBlanks(asStr, aiPos);
              {ignore trailing text}
    IF SysLocale.FarEast AND (System.Pos('ddd', ShortDateFormat) <> 0) THEN BEGIN
              {stop at time digit}
      IF ShortTimeFormat[1] IN ['0'..'9'] THEN BEGIN
        ScanToNumber(asStr, aiPos)
      END ELSE BEGIN
              {stop at time prefix}
        REPEAT
          WHILE (aiPos <= Length(asStr)) AND (asStr[aiPos] <> ' ') DO BEGIN
            Inc(aiPos);
          END {WHILE};

          ScanBlanks(asStr, aiPos);
        UNTIL (aiPos > Length(asStr))
        OR (AnsiCompareText(TimeAMString, Copy(asStr, aiPos, Length(TimeAMString))) = 0)
        OR (AnsiCompareText(TimePMString, Copy(asStr, aiPos, Length(TimePMString))) = 0);
      END {IF};
    END {IF};

    Result := DoEncodeDate(liYear, liMonth, liDay, afDate);
{}END {ScanDate};




{}FUNCTION    ScanTime(CONST asStr: STRING; VAR aiPos: INTEGER; VAR afTime: TDateTime): BOOLEAN;
{+H
---------------------------------------------------------------------------
  Remarks     - Same as the SysUtils procedure of the same name.
---------------------------------------------------------------------------}
  VAR
    liBaseHour: INTEGER;
    liHour    : WORD;
    liMinute  : WORD;
    liSecond  : WORD;
  BEGIN
    Result := FALSE;
    liBaseHour := -1;
    IF ScanString(asStr, aiPos, TimeAMString) OR ScanString(asStr, aiPos, 'AM') THEN BEGIN
      liBaseHour := 0
    END ELSE
    IF ScanString(asStr, aiPos, TimePMString) OR ScanString(asStr, aiPos, 'PM') THEN BEGIN
      liBaseHour := 12;
    END {IF};

    IF liBaseHour >= 0 THEN BEGIN
      ScanBlanks(asStr, aiPos);
    END {IF};

    IF NOT ScanNumber(asStr, aiPos, liHour) THEN BEGIN
      Exit;
    END {IF};

    liMinute := 0;
    IF ScanChar(asStr, aiPos, TimeSeparator) THEN BEGIN
      IF NOT ScanNumber(asStr, aiPos, liMinute) THEN BEGIN
        Exit;
      END {IF};
    END {IF};

    liSecond := 0;
    IF ScanChar(asStr, aiPos, TimeSeparator) THEN BEGIN
      IF NOT ScanNumber(asStr, aiPos, liSecond) THEN BEGIN
        Exit;
      END {IF};
    END {IF};

    IF liBaseHour < 0 THEN BEGIN
      IF ScanString(asStr, aiPos, TimeAMString) OR ScanString(asStr, aiPos, 'AM') THEN BEGIN
        liBaseHour := 0
      END ELSE
      IF ScanString(asStr, aiPos, TimePMString) OR ScanString(asStr, aiPos, 'PM') THEN BEGIN
        liBaseHour := 12;
      END {IF};
    END {IF};

    IF liBaseHour >= 0 THEN BEGIN
      IF (liHour = 0) OR (liHour > 12) THEN BEGIN
        Exit;
      END {IF};

      IF liHour = 12 THEN BEGIN
        liHour := 0;
      END {IF};

      Inc(liHour, liBaseHour);
    END {IF};

    ScanBlanks(asStr, aiPos);
    Result := DoEncodeTime(liHour, liMinute, liSecond, 0, afTime);
{}END {ScanTime};




{}FUNCTION    KSS_StrToDate(CONST asStr: STRING): TDateTime;
{+H
---------------------------------------------------------------------------
  Remarks     - Modified to handle date formats containing 3 character
                month names.
---------------------------------------------------------------------------}
  VAR
    leDateOrder    : TDateOrder;
    liPos          : INTEGER;
    ls             : STRING;
  BEGIN
    liPos := 1;
    ls := TranslateFormat(asStr,leDateOrder);
    IF ScanDate(ls, leDateOrder, liPos, Result) THEN BEGIN
      IF (liPos <= Length(ls)) THEN BEGIN
        ConvertErrorFmt(SInvalidDate, [ls]);
      END {IF};
    END ELSE BEGIN
      ConvertErrorFmt(SInvalidDate, [ls]);
    END {IF};
{}END {KSS_StrToDate};




{}FUNCTION    KSS_StrToDateTime(CONST asStr: STRING): TDateTime;
{+H
---------------------------------------------------------------------------
  Remarks     - Modified to handle date formats containing 3 character
                month names.
  Revised     - 1998.1120 (KSB) If not a date try to convert to a time value.
---------------------------------------------------------------------------}
  VAR
    leDateOrder    : TDateOrder;
    lfDate    : TDateTime;
    lfTime    : TDateTime;
    liPos     : INTEGER;
    ls        : STRING;
  BEGIN
    liPos  := 1;
    lfTime := 0;
    ls := TranslateFormat(asStr,leDateOrder);
    IF ScanDate(ls,leDateOrder,liPos,lfDate) THEN BEGIN
      IF (liPos > Length(ls)) THEN BEGIN
        lfTime := 0
      END ELSE
      IF NOT ScanTime(ls,liPos,lfTime) THEN BEGIN
        ConvertErrorFmt(SInvalidDateTime, [asStr]);
      END {IF};
    END ELSE BEGIN
      result := KSS_StrToTime(asStr);
      Exit;
    END {IF};

    IF lfDate >= 0 THEN BEGIN
      Result := lfDate + lfTime
    END ELSE BEGIN
      Result := lfDate - lfTime;
    END {IF};
{}END {KSS_StrToDateTime};




{}FUNCTION    KSS_StrToTime(CONST asStr: STRING): TDateTime;
{+H
---------------------------------------------------------------------------
  Remarks     - Modified to handle date formats containing 3 character
                month names.
---------------------------------------------------------------------------}
  VAR
    leDateOrder    : TDateOrder;
    liPos     : INTEGER;
    ls        : STRING;
  BEGIN
    liPos := 1;
    ls := TranslateFormat(asStr,leDateOrder);
    IF NOT ScanTime(ls, liPos, Result) OR (liPos <= Length(ls)) THEN BEGIN
      ConvertErrorFmt(SInvalidTime, [ls]);
    END {IF};
{}END {KSS_StrToTime};




END.
