UNIT KSSStore;
{+H
---------------------------------------------------------------------------
  File        - KSSSTORE.PAS
  Purpose     - Implements a multi-column, multi-row data store.
  Reference   - Based on TxtFile.pas (V:0.5) by Vladimir Savelev; ugms@sakhmail.sakhalin.ru
  Copyright    1998 Keith S. Brown, except as noted above.
  Remarks     - Implements a multi-column, multi-row data store that can contain
                string, double, date/time, integer-type, currency & boolean values.
                Its primary purposes is:
                1. read & store data from a delimited text file, where the first
                   row is assumed to be the column headings.
                2. store dynamically generated data.
                3. act as a data repository for KDataSet.

                Internally, all data is stored as strings. Non-string data is
                encoded hexadecimal coded strings. When the proper access methods
                & properties are used, this internal storage method is transparent
                to the programmer.

  Revised     - 1998.1027 (KSB) Wrote as its own unit.
              - 1998.1028 (KSB) Reorganized code, added TColumnDefs.
              - 1998.1103 (KSB) Added AllowBlankLines property.
              - 1998.1105 (KSB) Added SetStoreColumns procedure.
              - 1998.1106 (KSB) Corrected time conversion code. Added Report procedure.
              - 1998.1230 (KSB) Reversed Hexadecimal order.
---------------------------------------------------------------------------}
INTERFACE
USES
  Classes,  DB;


TYPE
  TTypeArray  = ARRAY[0..10000]OF TFieldType;
  PTypeArray  = ^TTypeArray;
  TTypeChangeEvent = PROCEDURE(aObject:TObject; aiCol:WORD; aeType:TFieldType) OF OBJECT;

              {TColumnDefs - stores names & widths of columns}
  TColumnDefs = CLASS(TStringList)
  PRIVATE
    FpTypes   : PTypeArray;
    FOnTypeChange  : TTypeChangeEvent;
    PROCEDURE   AllocateTypeMemory;
    PROCEDURE   DeallocateTypeMemory;
    PROCEDURE   CreateColumns(aiCol:WORD);
    FUNCTION    GetAsString(aiCol:WORD): STRING;
    FUNCTION    GetColName (aiCol:WORD): STRING;
    FUNCTION    GetColType (aiCol:WORD): TFieldType;
    FUNCTION    GetColWidth(aiCol:WORD): WORD;
    PROCEDURE   SetAsString(aiCol:WORD; asDef:STRING);
    PROCEDURE   SetColName (aiCol:WORD; asName:STRING);
    PROCEDURE   SetColType (aiCol:WORD; aeType:TFieldType);
    PROCEDURE   SetColWidth(aiCol:WORD; aiWidth:WORD);
  PUBLIC
    PROCEDURE   CopyFrom(aoSource:TPersistent); 
    PROCEDURE   Clear; OVERRIDE;
    CONSTRUCTOR Create;
    DESTRUCTOR  Destroy; OVERRIDE;
  PUBLIC
    CLASS FUNCTION  ColumnTypeToStr(aeDataType:TFieldType):STRING;
    CLASS PROCEDURE DecodeColumnStr(asCol:STRING; VAR asColName:STRING; VAR aeDataType:TFieldType; VAR aiWidth:WORD);
    CLASS FUNCTION  EncodeColumnStr(asColName:STRING; aeDataType:TFieldType; aiWidth:WORD):STRING;
    CLASS FUNCTION  StrToColumnType(asDataType:STRING):TFieldType;
  PUBLIC
    PROPERTY    AsString   [aiCol:WORD]: STRING     READ GetAsString WRITE SetAsString;
    PROPERTY    ColumnName [aiCol:WORD]: STRING     READ GetColName  WRITE SetColName;
    PROPERTY    ColumnType [aiCol:WORD]: TFieldType READ GetColType  WRITE SetColType;
    PROPERTY    ColumnWidth[aiCol:WORD]: WORD       READ GetColWidth WRITE SetColWidth;
  PUBLIC
    PROPERTY    OnTypeChange:TTypeChangeEvent READ FOnTypeChange WRITE FOnTypeChange;
  END {CLASS};

              {TDataStore - keeps data in memory.
               Also implements methods for retrieving and storing data in text files}
  TDataStore  = CLASS(TStringList)
  PRIVATE
    FbBlankLines   : BOOLEAN;
    FcDelimt       : CHAR;
    FiCol          : LongINT;
    FiRow          : LongINT;
    FoNames        : TColumnDefs;
    FOnTypeChange  : TTypeChangeEvent;
    FUNCTION    GetCell    (aiCol, aiRow: WORD): STRING;
    FUNCTION    GetCObject (aiCol:WORD): TObject;
    FUNCTION    GetRObject (aiRow:WORD): TObject;
    PROCEDURE   PutCObject (aiCol:WORD; CONST aoObject:TObject);
    PROCEDURE   PutRObject (aiRow:WORD; CONST aoObject:TObject);
    PROCEDURE   ReadRow    (aoReader:TReader);
    PROCEDURE   ReadCol    (aoReader:TReader);
    PROCEDURE   SetCell    (aiCol,aiRow:WORD; CONST asValue:STRING);
    PROCEDURE   WriteCol   (aoWriter:TWriter);
    PROCEDURE   WriteRow   (aoWriter:TWriter);
  PROTECTED
    PROCEDURE   ChangeColumnType(aoSender:TObject; aiCol:WORD; aeType:TFieldType); VIRTUAL;
    PROCEDURE   DefineProperties(aoFiler:TFiler); OVERRIDE;
    FUNCTION    GetCell_AsString(aiCol, aiRow:WORD): STRING;           VIRTUAL;
    FUNCTION    GetCell_Boolean (aiCol, aiRow:WORD): Boolean;          VIRTUAL;
    FUNCTION    GetCell_Currency(aiCol, aiRow:WORD): Currency;         VIRTUAL;
    FUNCTION    GetCell_DateTime(aiCol, aiRow:WORD): TDateTime;        VIRTUAL;
    FUNCTION    GetCell_Float   (aiCol, aiRow:WORD): Double;           VIRTUAL;
    FUNCTION    GetCell_Integer (aiCol, aiRow:WORD): Longint;          VIRTUAL;
    PROCEDURE   SetCell_AsString(aiCol, aiRow:WORD; aValue:STRING);    VIRTUAL;
    PROCEDURE   SetCell_Boolean (aiCol, aiRow:WORD; aValue:BOOLEAN);   VIRTUAL;
    PROCEDURE   SetCell_Currency(aiCol, aiRow:WORD; aValue:CURRENCY);  VIRTUAL;
    PROCEDURE   SetCell_Date    (aiCol, aiRow:WORD; aValue:TDateTime); VIRTUAL;
    PROCEDURE   SetCell_DateTime(aiCol, aiRow:WORD; aValue:TDateTime); VIRTUAL;
    PROCEDURE   SetCell_Float   (aiCol, aiRow:WORD; aValue:DOUBLE);    VIRTUAL;
    PROCEDURE   SetCell_Integer (aiCol, aiRow:WORD; aValue:LongINT);   VIRTUAL;
    PROCEDURE   SetCell_SmallInt(aiCol, aiRow:WORD; aValue:LongINT);   VIRTUAL;
    PROCEDURE   SetCell_Time    (aiCol, aiRow:WORD; aValue:TDateTime); VIRTUAL;
    PROCEDURE   SetCell_Word    (aiCol, aiRow:WORD; aValue:LongINT);   VIRTUAL;
  PROTECTED
    PROCEDURE   InsertRow(aiRow:WORD; asStr:STRING; aoObj:TObject);
  PROTECTED
    PROPERTY    Cell        [aiCol,aiRow: WORD]:STRING    READ GetCell          WRITE SetCell; DEFAULT;
  PUBLIC
    PROCEDURE   AddRow(asStr:STRING; aoObj:TObject);
    PROCEDURE   Clear; OVERRIDE;
    CONSTRUCTOR Create;
    PROCEDURE   DeleteRow(aiRow:WORD);
    DESTRUCTOR  Destroy; OVERRIDE;
    PROCEDURE   ClearData;
    FUNCTION    IndexOfObject(CONST aoObject:TObject):LongINT;
    PROCEDURE   LoadFromFile(CONST asFileName:STRING); OVERRIDE;
    PROCEDURE   SaveToFile  (CONST asFileName:STRING); OVERRIDE;
    PROCEDURE   SetStoreColumns(aiColumns:WORD);
  PUBLIC
    PROPERTY    BooleanCell [aiCol,aiRow: WORD]:BOOLEAN   READ GetCell_Boolean  WRITE SetCell_Boolean;
    PROPERTY    CellAsString[aiCol,aiRow: WORD]:STRING    READ GetCell_AsString WRITE SetCell_AsString;
    PROPERTY    CurrencyCell[aiCol,aiRow: WORD]:CURRENCY  READ GetCell_Currency WRITE SetCell_Currency;
    PROPERTY    DateCell    [aiCol,aiRow: WORD]:TDateTime READ GetCell_DateTime WRITE SetCell_Date;
    PROPERTY    DateTimeCell[aiCol,aiRow: WORD]:TDateTime READ GetCell_DateTime WRITE SetCell_DateTime;
    PROPERTY    FloatCell   [aiCol,aiRow: WORD]:DOUBLE    READ GetCell_Float    WRITE SetCell_Float;
    PROPERTY    IntegerCell [aiCol,aiRow: WORD]:LongINT   READ GetCell_Integer  WRITE SetCell_Integer;
    PROPERTY    SmallIntCell[aiCol,aiRow: WORD]:LongINT   READ GetCell_Integer  WRITE SetCell_SmallInt;
    PROPERTY    TimeCell    [aiCol,aiRow: WORD]:TDateTime READ GetCell_DateTime WRITE SetCell_Time;
    PROPERTY    WordCell    [aiCol,aiRow: WORD]:LongINT   READ GetCell_Integer  WRITE SetCell_Word;
  PUBLIC
    PROPERTY    ColObjects [aiCol: WORD] : TObject READ GetCObject  WRITE PutCObject;
    PROPERTY    RowObjects [aiRow: WORD] : TObject READ GetRObject  WRITE PutRObject;
  PUBLIC
    PROPERTY    ColCount : LongINT     READ FiCol;
    PROPERTY    Delimiter: CHAR        READ FcDelimt    WRITE FcDelimt;
    PROPERTY    Header   : TColumnDefs READ FoNames     WRITE FoNames;
    PROPERTY    RowCount : LongINT     READ FiRow;
  PUBLISHED
    PROPERTY    AllowBlankLines:BOOLEAN       READ FbBlankLines  WRITE FbBlankLines;
    PROPERTY    OnTypeChange:TTypeChangeEvent READ FOnTypeChange WRITE FOnTypeChange;
  END {CLASS};


{}PROCEDURE   ColumnCreateError(aiCol:WORD; aeType:TFieldType);
{}FUNCTION    DefaultDelimiter:CHAR;
              {=================================================}
IMPLEMENTATION
USES
  SysUtils, KEnums,   KSSDates;


CONST
  ccDefaultDelimit : CHAR = #9;
CONST
  cAcceptableTypes = [ftBoolean,ftCurrency,ftDate,ftDateTime,ftFloat,ftInteger,ftSmallInt,ftString,ftTime,ftWord];
  ciMaxRecLength   = 4096;
TYPE
  T2Bytes     = ARRAY[0..1]OF BYTE;
  T4Bytes     = ARRAY[0..3]OF BYTE;
  T8Bytes     = ARRAY[0..7]OF BYTE;

  EConvertErr = CLASS(Exception);
  ERangeError = CLASS(Exception);
  EInvalidType= CLASS(Exception);


{}PROCEDURE   BadSmallIntValue(aiValue:INTEGER);
  BEGIN
    RAISE EConvertErr.Create(IntToStr(aiValue)+' is not a legal value for SmallInt');
{}END {BadSmallIntValue};




{}PROCEDURE   BadWordValue(aiValue:INTEGER);
  BEGIN
    RAISE EConvertErr.Create(IntToStr(aiValue)+' is not a legal value for word');
{}END {BadWordValue};




{}PROCEDURE   ColumnConvertError(aiCol:WORD; aeType:TFieldType);
  VAR
    loEnums   : TKustomEnum;
    ls        : STRING;
  BEGIN
    loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
    ls := loEnums.ToString(Ord(aeType))+'.';
    loEnums.Free;
    RAISE EConvertErr.Create('Cannot convert column '+IntToStr(aiCol)+' to '+ls);
{}END {ColumnConvertError};




{}PROCEDURE   ColumnCreateError(aiCol:WORD; aeType:TFieldType);
  VAR
    loEnums   : TKustomEnum;
    ls        : STRING;
  BEGIN
    loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
    ls := loEnums.ToString(Ord(aeType))+'.';
    loEnums.Free;
    RAISE EInvalidType.Create('Error in column '+IntToStr(aiCol)+'. Cannot create columns of type '+ls);
{}END {ColumnCreateError};




{}PROCEDURE   ColumnTypeIsWrong(aiCol:WORD; aeType:TFieldType);
  VAR
    loEnums   : TKustomEnum;
    ls        : STRING;
  BEGIN
    loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
    ls := loEnums.ToString(Ord(aeType));
    loEnums.Free;
    RAISE EInvalidType.Create('Column '+IntToStr(aiCol)+' is not a '+ls+' column.');
{}END {ColumnTypeIsWrong};




{}PROCEDURE   TypeError(aeType:TFieldType);
  VAR
    loEnums   : TKustomEnum;
    ls        : STRING;
  BEGIN
    loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
    ls := loEnums.ToString(Ord(aeType));
    loEnums.Free;
    RAISE EInvalidType.Create('Cannot create or use '+ls+' type data');
{}END {TypeError};




{}FUNCTION    DefaultDelimiter:CHAR;
  BEGIN
    result := ccDefaultDelimit;
{}END {DefaultDelimiter};




{}FUNCTION    HexStrToByte(aHexValue:STRING):BYTE;
{+H
---------------------------------------------------------------------------
  Purpose     - returns a byte value for a string that consists of two
                hexidecimal digits.
---------------------------------------------------------------------------}
  VAR
    li        : BYTE;
    liCode    : INTEGER;
    ls        : STRING[2];
  BEGIN
    ls := aHexValue;
    System.Val('$'+ls,li,liCode);
    result := li;
{}END {HexStrToByte};




{}FUNCTION    IntTo4ByteHexStr(aiLong:LongINT):STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Return an 8 character hexadecimal coded string for a 4 byte
                integer value.
  Revised     - 1998.1230 (KSB) Reversed order.
---------------------------------------------------------------------------}
{}{}FUNCTION    _4_BytesToHexStr(aaBytes:T4Bytes):STRING;
    BEGIN
      result := SysUtils.IntToHex(aaBytes[3],2)
              + SysUtils.IntToHex(aaBytes[2],2)
              + SysUtils.IntToHex(aaBytes[1],2)
              + SysUtils.IntToHex(aaBytes[0],2)
{}{}END {_4_BytesToHexStr};


  VAR
    laBytes   : T4Bytes ABSOLUTE aiLong;
  BEGIN
    result := _4_BytesToHexStr(laBytes);
{}END {IntTo4ByteHexStr};




{}FUNCTION    _8_BytesToHexStr(aaBytes:T8Bytes):STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Return a 16 character hexadecimal coded string for 8 bytes
                of binary data.
  Revised     - 1998.1230 (KSB) Reversed order.
---------------------------------------------------------------------------}
  BEGIN
    result := SysUtils.IntToHex(aaBytes[7],2)
            + SysUtils.IntToHex(aaBytes[6],2)
            + SysUtils.IntToHex(aaBytes[5],2)
            + SysUtils.IntToHex(aaBytes[4],2)
            + SysUtils.IntToHex(aaBytes[3],2)
            + SysUtils.IntToHex(aaBytes[2],2)
            + SysUtils.IntToHex(aaBytes[1],2)
            + SysUtils.IntToHex(aaBytes[0],2)
{}END {_8_BytesToHexStr};




{}FUNCTION    _4ByteHexStrToInt(asHexStr:STRING):LongINT;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert a maximum of 8 Hexadecimal characters to a 1, 2, 4
                byte binary (integer) value.
  Revised     - 1998.1230 (KSB) Reversed order.
---------------------------------------------------------------------------}
{}{}FUNCTION    ConvertTo_4_Bytes(asHexStr:STRING):T4Bytes;
    BEGIN
      FillChar(result,SizeOf(T4Bytes),0);
      IF Length(asHexStr) = 8 THEN BEGIN
        result[3] := HexStrToByte(Copy(asHexStr,01,2));
        result[2] := HexStrToByte(Copy(asHexStr,03,2));
        result[1] := HexStrToByte(Copy(asHexStr,05,2));
        result[0] := HexStrToByte(Copy(asHexStr,07,2));
      END ELSE BEGIN
        RAISE EConvertErr.Create('"'+asHexStr+'" is not a 4 byte hex string.');
      END {IF};
{}{}END {ConvertTo_4_Bytes};


  VAR
    li        : LongINT;
    laBytes   : T4Bytes ABSOLUTE li;
  BEGIN
    laBytes := ConvertTo_4_Bytes(asHexStr);
    result  := li;
{}END {_4ByteHexStrToInt};




{}FUNCTION    ConvertTo_8_Bytes(asHexStr:STRING):T8Bytes;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert 16 Hexadecimal characters to 8 bytes of binary data.
  Revised     - 1998.1230 (KSB) Reversed order.
---------------------------------------------------------------------------}
  BEGIN
    IF Length(asHexStr) = 16 THEN BEGIN
      result[7] := HexStrToByte(Copy(asHexStr,01,2));
      result[6] := HexStrToByte(Copy(asHexStr,03,2));
      result[5] := HexStrToByte(Copy(asHexStr,05,2));
      result[4] := HexStrToByte(Copy(asHexStr,07,2));
      result[3] := HexStrToByte(Copy(asHexStr,09,2));
      result[2] := HexStrToByte(Copy(asHexStr,11,2));
      result[1] := HexStrToByte(Copy(asHexStr,13,2));
      result[0] := HexStrToByte(Copy(asHexStr,15,2));
    END ELSE BEGIN
      RAISE EConvertErr.Create('"'+asHexStr+'" is not a 8 byte hex string.');
    END {IF};
{}END {ConvertTo_8_Bytes};




{}FUNCTION    _8ByteHexStrToCurrency(asHexStr:STRING):CURRENCY;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert 16 Hexadecimal characters to an 8 byte currency value.
---------------------------------------------------------------------------}
  VAR
    lf        : CURRENCY;
    laBytes   : T8Bytes ABSOLUTE lf;
  BEGIN
    laBytes := ConvertTo_8_Bytes(asHexStr);
    result  := lf;
{}END {_8ByteHexStrToCurrency};




{}FUNCTION    _8ByteHexStrToFloat(asHexStr:STRING):DOUBLE;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert 16 Hexadecimal characters to an 8 byte Double value.
---------------------------------------------------------------------------}
  VAR
    lf        : DOUBLE;
    laBytes   : T8Bytes ABSOLUTE lf;
  BEGIN
    laBytes := ConvertTo_8_Bytes(asHexStr);
    result  := lf;
{}END {_8ByteHexStrToFloat};




{}FUNCTION    CurrencyTo8ByteHexStr(afValue:CURRENCY):STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert a currency value to a string of 16 Hexadecimal
                characters.
---------------------------------------------------------------------------}
  VAR
    laBytes   : T8Bytes ABSOLUTE afValue;
  BEGIN
    result := _8_BytesToHexStr(laBytes);
{}END {CurrencyTo8ByteHexStr};




{}FUNCTION    DoubleTo8ByteHexStr(afValue:DOUBLE):STRING;
{+H
---------------------------------------------------------------------------
  Purpose     - Convert a double precision floating point value to a string
                of 16 Hexadecimal characters.
---------------------------------------------------------------------------}
  VAR
    laBytes   : T8Bytes ABSOLUTE afValue;
  BEGIN
    result := _8_BytesToHexStr(laBytes);
{}END {DoubleTo8ByteHexStr};




{}FUNCTION    StrToCurrency(asStr:STRING):CURRENCY;
{+H
---------------------------------------------------------------------------
  Purpose     - Given a string representation return its currency value.
---------------------------------------------------------------------------}
  VAR
    lbNeg     : BOOLEAN;
    ls        : STRING;
    lsCurrency: STRING;
  BEGIN
    lbNeg := FALSE;
          {copy the string used to indicate monetary units}
    lsCurrency := CurrencyString;
          {strip the formatting from "asStr" and build
           in "ls" an equivalent fixed point string}
    ls := '';
    WHILE Length(asStr) > 0 DO BEGIN
      CASE asStr[1] OF
        '0'..'9' : ls := ls + asStr[1];
        ELSE BEGIN
          IF asStr[1] IN ['-','('] THEN BEGIN
            lbNeg := TRUE;
          END ELSE
          IF (Length(lsCurrency) > 0) AND (asStr[1] = lsCurrency[1]) THEN BEGIN
            System.Delete(lsCurrency,1,1);
          END ELSE
          IF asStr[1] = ThousandSeparator THEN BEGIN
            {do nothing}
          END ELSE
          IF asStr[1] = DecimalSeparator THEN BEGIN
            ls := ls + '.';
          END {IF};
        END {ELSE};
      END {CASE};

      System.Delete(asStr,1,1);
    END {WHILE};

          {convert the fixed point string to a fixed point number}
    result := StrToFloat(ls);
    IF lbNeg THEN BEGIN
      result := -result;
    END {IF};
{}END {StrToCurrency};


              {TColumnDefs Methods}


{}PROCEDURE   TColumnDefs.CopyFrom(aoSource:TPersistent);
{+O
---------------------------------------------------------------------------
  Purpose     - Copy information from either a stringlist or from a
                different TColumnDef object to this TColumnDef object.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
    lp        : PTypeArray;
  BEGIN
    FOR li := 1 TO (aoSource AS TStringList).Count DO BEGIN
      IF li <= Count THEN BEGIN
        Strings[li-1] := (aoSource AS TStringList).Strings[li-1];
        Objects[li-1] := (aoSource AS TStringList).Objects[li-1];
      END ELSE BEGIN
        AddObject(TStringList(aoSource)[li-1],TStringList(aoSource).Objects[li-1]);
      END {IF};
    END {FOR};

    IF (aoSource IS TColumnDefs) THEN BEGIN
      FOR li := 1 TO TColumnDefs(aoSource).Count DO BEGIN
        ColumnType[li-1] := TColumnDefs(aoSource).ColumnType[li-1];
      END {FOR};
    END {IF};

    IF Count > (aoSource AS TStringList).Count THEN BEGIN
      IF FpTypes <> NIL THEN BEGIN
        li := (aoSource AS TStringList).Count;
        GetMem(lp, SizeOf(TFieldType)*li);
        System.Move(TTypeArray(FpTypes^), TTypeArray(lp^), SizeOf(TFieldType)*li);
        FreeMem(FpTypes);
        FpTypes := lp;
      END {IF};

      li := Count;
      WHILE (li > (aoSource AS TStringList).Count) DO BEGIN
        Dec(li);
        Delete(li);
      END {WHILE};
    END {IF};
{}END {CopyFrom};




{}PROCEDURE   TColumnDefs.Clear;
{+O
---------------------------------------------------------------------------
  Purpose     - Remove all data from this TColumnDef object.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Clear;
    DeallocateTypeMemory
{}END {Clear};




{}CONSTRUCTOR TColumnDefs.Create;
{+O
---------------------------------------------------------------------------
  Purpose     - Create this TColumnDef object.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Create;
    FpTypes  := NIL;
{}END {Create};




{}PROCEDURE   TColumnDefs.AllocateTypeMemory;
{+O
---------------------------------------------------------------------------
  Purpose     - Allocate memory to hold column types.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    IF FpTypes = NIL THEN BEGIN
      GetMem( FpTypes, SizeOf(TFieldType)*Count );
      FOR li := 0 TO Pred(Count) DO BEGIN
        FpTypes^[li] := ftString;
      END {FOR};
    END {IF};
{}END {AllocateTypeMemory};




{}PROCEDURE   TColumnDefs.CreateColumns(aiCol:WORD);
{+O
---------------------------------------------------------------------------
  Purpose     - Ensure that the appropriate number of columns exist -- and
                if the identified column does not exist, then create it.
  Revised     - 1998.1105 (KSB) Fixed bugs.
              - 1998.1228 (KSB) Changed control logic.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
    lp        : PTypeArray;
    ls        : STRING;
  BEGIN
    IF aiCol > Count-1 THEN BEGIN
      IF FpTypes <> NIL THEN BEGIN
        GetMem(lp, SizeOf(TFieldType)*aiCol);
        System.Move(TTypeArray(FpTypes^), TTypeArray(lp^), SizeOf(TFieldType)*Count);
        FreeMem(FpTypes);
        FpTypes := lp;
              {1998.1105 Give the new type slot a default value}
        FpTypes[aiCol] := ftString;
      END {IF};

              {1998.1105 changd from aiCol-1 to aiCol}
              {provide a default name & size for the new slot(s)}
      FOR li := Count TO aiCol DO BEGIN
        ls := 'Column '+IntToStr(li);
        AddObject(ls,POINTER(Length(ls)+5));
      END {FOR};

      IF FpTypes = NIL THEN BEGIN
        AllocateTypeMemory;
      END {IF};
    END {IF};
{}END {CreateColumns};




{}PROCEDURE   TColumnDefs.DeallocateTypeMemory;
{+O
---------------------------------------------------------------------------
  Purpose     - Free memory previously allocated to hold column types.
---------------------------------------------------------------------------}
  BEGIN
    IF FpTypes <> NIL THEN BEGIN
      FreeMem( FpTypes );
      FpTypes := NIL;
    END {IF};
{}END {DeallocateTypeMemory};




{}DESTRUCTOR  TColumnDefs.Destroy;
{+O
---------------------------------------------------------------------------
  Purpose     - Remove the column type information if it exists.
---------------------------------------------------------------------------}
  BEGIN
    DeallocateTypeMemory;
    INHERITED Destroy;
{}END {Destroy};




{}FUNCTION    TColumnDefs.GetAsString(aiCol:WORD):STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a string representing the column name, its type & width.
---------------------------------------------------------------------------}
  BEGIN
    result := EncodeColumnStr(GetColName(aiCol),GetColType(aiCol),GetColWidth(aiCol));
{}END {GetAsString};




{}FUNCTION    TColumnDefs.GetColName(aiCol:WORD):STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a string containing the name of the column.
---------------------------------------------------------------------------}
  BEGIN
    CreateColumns(aiCol);
    result := self[aiCol];
{}END {GetColName};




{}FUNCTION    TColumnDefs.GetColType(aiCol:WORD):TFieldType;
{+O
---------------------------------------------------------------------------
  Purpose     - Return an enumerated value for the column type.
---------------------------------------------------------------------------}
  BEGIN
    IF FpTypes = NIL THEN BEGIN
      result := ftString;
    END ELSE BEGIN
      CreateColumns(aiCol);
      result := FpTypes^[aiCol];
    END {IF};
{}END {GetColType};




{}FUNCTION    TColumnDefs.GetColWidth(aiCol:WORD):WORD;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a unsigned 2 byte integer for the width of the column.
---------------------------------------------------------------------------}
  BEGIN
    CreateColumns(aiCol);
    result := INTEGER(Objects[aiCol])
{}END {GetColWidth};




{}PROCEDURE   TColumnDefs.SetAsString(aiCol:WORD; asDef:STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Set the name, type and width of column "aiCol" to the values
                contained in the string "asDef"
---------------------------------------------------------------------------}
  VAR
    leType    : TFieldType;
    liWidth   : WORD;
    lsName    : STRING;
  BEGIN
    DecodeColumnStr(asDef,lsName,leType,liWidth);
    SetColName(aiCol,lsName);
    SetColType(aiCol,leType);
    SetColWidth(aiCol,liWidth);
{}END {SetAsString};




{}PROCEDURE   TColumnDefs.SetColName(aiCol:WORD; asName:STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Set the name of column "aiCol" to "asName"
---------------------------------------------------------------------------}
  BEGIN
    CreateColumns(aiCol);
    Strings[aiCol] := asName;
{}END {SetColName};




{}PROCEDURE   TColumnDefs.SetColType(aiCol:WORD; aeType:TFieldType);
{+O
---------------------------------------------------------------------------
  Purpose     - Change the data type for column "aiCol" to "aeType".
---------------------------------------------------------------------------}
  BEGIN
    CreateColumns(aiCol);
    IF FpTypes = NIL THEN BEGIN
      AllocateTypeMemory;
    END {IF};

    IF Assigned(FOnTypeChange) THEN BEGIN
      FOnTypeChange(self,aiCol,aeType);
    END {IF};

    FpTypes^[aiCol] := aeType;
{}END {SetColType};




{}PROCEDURE   TColumnDefs.SetColWidth(aiCol:WORD; aiWidth:WORD);
{+O
---------------------------------------------------------------------------
  Purpose     - Change the width of column "aiCol" to "aiWidth" characters.
---------------------------------------------------------------------------}
  BEGIN
    CreateColumns(aiCol);
    objects[aiCol] := POINTER(aiWidth);
{}END {SetColWidth};


              {KSS TColumnDefs Class Methods}


{}CLASS FUNCTION  TColumnDefs.ColumnTypeToStr(aeDataType:TFieldType):STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a string representing the enumerated data type "aeDataType".
---------------------------------------------------------------------------}
  VAR
    loEnums   : TKustomEnum;
  BEGIN
    IF aeDataType IN cAcceptableTypes THEN BEGIN
      loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
      result := loEnums.ToString(Ord(aeDataType));
      loEnums.Free;
    END ELSE BEGIN
//    TypeError(aeDataType);
      result := 'ftString';
    END {IF};
{}END {ColumnTypeToStr};




{}CLASS PROCEDURE TColumnDefs.DecodeColumnStr(asCol:STRING; VAR asColName:STRING; VAR aeDataType:TFieldType; VAR aiWidth:WORD);
{+O
---------------------------------------------------------------------------
  Purpose     - Given the string "asCol" containing a column name, type &
                width, return those values.
---------------------------------------------------------------------------}
  CONST
    cWhiteSpace    = [',',';'];
  TYPE
    TePortion = (_name,_type,_width);
  VAR
    ls        : STRING;
    le        : TePortion;
    li        : INTEGER;
    liStart   : INTEGER;
  BEGIN
    asColName := '';
    aeDataType:= ftString;
    asCol  := Trim(asCol) + ',';
    le     := _name;
    li     := 1;
    liStart:= 1;
    WHILE li <= Length(asCol) DO BEGIN
      IF asCol[li] IN cWhiteSpace THEN BEGIN
        ls := Copy(asCol,liStart,li-liStart);
        CASE le OF
          _name : BEGIN
            asColName := Trim(ls);
            le := _type;
          END {_name};

          _type  : BEGIN
            aeDataType := StrToColumnType(Trim(ls));
            le := _width;
          END {_type};

          _width : BEGIN
            aiWidth := StrToIntDef(Trim(ls),15);
            Exit;
          END {_width};
        END {CASE};

        WHILE (li < Length(asCol)) AND (asCol[li] IN cWhiteSpace) DO BEGIN
          Inc(li);
        END {WHILE};

        liStart := li;
      END {IF};

      Inc(li);
    END {WHILE};
{}END {DecodeColumnStr};




{}CLASS FUNCTION  TColumnDefs.EncodeColumnStr(asColName:STRING; aeDataType:TFieldType; aiWidth:WORD):STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a string representing a name, data type & width.
---------------------------------------------------------------------------}
  BEGIN
    result := asColName+','+ColumnTypeToStr(aeDataType)+','+IntToStr(aiWidth);
{}END {EncodeColumnStr};




{}CLASS FUNCTION  TColumnDefs.StrToColumnType(asDataType:STRING):TFieldType;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a data type for its string representation.
---------------------------------------------------------------------------}
  VAR
    le        : TFieldType;
    loEnums   : TKustomEnum;
  BEGIN
    asDataType := UpperCase(asDataType);
    loEnums := TKustomEnum.Create(TypeInfo(TFieldType));
    result  := ftBlob;
    TRY
      FOR le := TFieldType(0) TO High(TFieldType) DO BEGIN
        IF UpperCase(loEnums.ToString(Ord(le))) = asDataType THEN BEGIN
          result := le;
          Break;
        END {IF};
      END {FOR};
    FINALLY
      loEnums.Free;
    END {TRY};

    IF result IN cAcceptableTypes THEN BEGIN
      {type "ok"}
    END ELSE BEGIN
//    TypeError(result);
      result := ftString;
    END {IF};
{}END {StrToColumnType};


              {KSS Data Store Methods}


{}PROCEDURE   TDataStore.AddRow(asStr:STRING; aoObj:TObject);
{+O
---------------------------------------------------------------------------
  Purpose     - Append a new row to the data store
  Revised     - 1998.1116 (KSB) Wrote initial version.
---------------------------------------------------------------------------}
  BEGIN
    InsertRow(RowCount,asStr,aoObj);
{}END {AddRow};




{}PROCEDURE   TDataStore.ChangeColumnType(aoSender:TObject; aiCol:WORD; aeType:TFieldType);
{+O
---------------------------------------------------------------------------
  Purpose     - Change the data type for a column in the data store.
  Remarks     - The column type in the header is set to aeType _after_ this
                procedure executes...
  Revised     - 1998.1106 (KSB) Corrected time conversion.
---------------------------------------------------------------------------}
{}{}PROCEDURE StrToHexBool;
    VAR
      liRow   : INTEGER;
      ls      : STRING;
    BEGIN
      FOR liRow := 0 TO RowCount-1 DO BEGIN
        ls := UpperCase(Cell[aiCol,liRow]);
        IF (ls = 'TRUE') OR (ls = 'YES') THEN BEGIN
          Cell[aiCol,liRow] := IntTo4ByteHexStr(Ord(TRUE));
        END ELSE BEGIN
          Cell[aiCol,liRow] := IntTo4ByteHexStr(Ord(FALSE));
        END {IF};
      END {FOR};
{}{}END {StrToHexBool};


{}{}PROCEDURE StrToHexDateTime;
    VAR
      lfDT    : TDateTime;
      liRow   : INTEGER;
      ls      : STRING;
    BEGIN
      FOR liRow := 0 TO RowCount-1 DO BEGIN
        ls := Cell[aiCol,liRow];
        IF ls = '' THEN BEGIN
          lfDT := 0;
        END ELSE BEGIN
          CASE aeType OF
            ftDate,
            ftDateTime : lfDT := KSSDates.KSS_StrToDateTime(ls);
            ftTime     : lfDT := KSSDates.KSS_StrToTime(ls);
            ELSE         lfDT := 0;
          END {CASE};
        END {IF};

        Cell[aiCol,liRow] := DoubleTo8ByteHexStr(lfDT);
      END {FOR};
{}{}END {StrToHexDateTime};


{}{}PROCEDURE StrToHexDollars;
    VAR
      lfDollar: CURRENCY;
      liRow   : INTEGER;
      ls      : STRING;
    BEGIN
      FOR liRow := 0 TO RowCount-1 DO BEGIN
        ls := UpperCase(Cell[aiCol,liRow]);
        IF ls = '' THEN BEGIN
          lfDollar := 0;
        END ELSE BEGIN
          lfDollar := StrToCurrency(ls);
        END {IF};

        Cell[aiCol,liRow] := CurrencyTo8ByteHexStr(lfDollar);
      END {FOR};
{}{}END {StrToHexDollars};


{}{}PROCEDURE StrToHexFloat;
    VAR
      lfD     : TDateTime;
      liRow   : INTEGER;
      ls      : STRING;
    BEGIN
      FOR liRow := 0 TO RowCount-1 DO BEGIN
        ls  := Cell[aiCol,liRow];
        IF ls = '' THEN BEGIN
          lfD := 0.0;
        END ELSE BEGIN
          lfD := SysUtils.StrToFloat(ls);
        END {IF};

        Cell[aiCol,liRow] := DoubleTo8ByteHexStr(lfD);
      END {FOR};
{}{}END {StrToHexFloat};


{}{}PROCEDURE StrToHexInt;
    VAR
      li      : INTEGER;
      liRow   : INTEGER;
    BEGIN
      FOR liRow := 0 TO RowCount-1 DO BEGIN
        li := SysUtils.StrToIntDef(Cell[aiCol,liRow],0);
        CASE Header.ColumnType[aiCol] OF
          ftWord     : BEGIN
            IF (li < 0) OR (li > 65535) THEN BEGIN
              BadWordValue(li);
            END {IF};
          END {ftWord};

          ftSmallInt : BEGIN
            IF (li < -32768) OR (li > 32767) THEN BEGIN
              BadSmallIntValue(li);
            END {IF};
          END {ftSmallInt};

          ftInteger  :;
        END {CASE};

        Cell[aiCol,liRow] := IntTo4ByteHexStr(li);
      END {FOR};
{}{}END {StrToHexInt};


  VAR
    leOldType : TFieldType;
    liRow     : INTEGER;
  BEGIN
    leOldType := Header.ColumnType[aiCol];
    CASE aeType OF
      ftString : BEGIN
        FOR liRow := 0 TO RowCount-1 DO BEGIN
          Cell[aiCol,liRow] := GetCell_AsString(aiCol,liRow);
        END {FOR};
      END {ftString};

      ftWord : BEGIN
        CASE leOldType OF
          ftString : StrToHexInt;
          ftWord   :;   {no change}
          ELSE       ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftWord};

      ftSmallint : BEGIN
        CASE leOldType OF
          ftString  : StrToHexInt;
          ftSmallInt:;  {no change}
          ELSE        ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftSmallint};

      ftInteger : BEGIN
        CASE leOldType OF
          ftString  : StrToHexInt;
          ftInteger :;  {no change}
          ftWord    :;  {no change}
          ftSmallint:;  {no change}
          ELSE        ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftInteger};

      ftBoolean : BEGIN
        CASE leOldType OF
          ftString : StrToHexBool;
          ftBoolean:;   {no change}
          ELSE       ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftBoolean};

      ftFloat : BEGIN
        CASE leOldType OF
          ftString : StrToHexFloat;
          ftFloat  :;   {no change}
          ELSE       ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftFloat};

      ftCurrency : BEGIN
        CASE leOldType OF
          ftString  : StrToHexDollars;
          ftCurrency:;  {no change}
          ELSE        ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftCurrency};

      ftDate : BEGIN
        CASE leOldType OF
          ftString : StrToHexDateTime;
          ftDate   :;   {no change}
          ELSE       ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftDate};

      ftTime : BEGIN
        CASE leOldType OF
          ftString : StrToHexDateTime;
          ftTime   :;   {no change}
          ELSE       ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftTime};

      ftDateTime : BEGIN
        CASE leOldType OF
          ftString  : StrToHexDateTime;
          ftDateTime:;  {no change}
          ftDate    :;  {no change}
          ftTime    :;  {no change}
          ELSE        ColumnConvertError(aiCol,aeType);
        END {CASE};
      END {ftDateTime};

      ELSE BEGIN
        ColumnCreateError(aiCol,aeType);
       {ftUnknown, ftBCD,     ftBytes,   ftVarBytes,   ftAutoInc,   ftBlob,
        ftMemo,    ftGraphic, ftFmtMemo, ftParadoxOle, ftDBaseOle,  ftTypedBinary,
        ftCursor}
      END {ELSE};
    END {CASE};
{}END {ChangeColumnType};




{}PROCEDURE   TDataStore.Clear;
{+O
---------------------------------------------------------------------------
  Purpose     - Empty the data store & remove the column type information if
                it exists.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Clear;
{}END {Clear};




{}PROCEDURE   TDataStore.ClearData;
{+O
---------------------------------------------------------------------------
  Purpose     - Remove data from the data store leaving one empty row (row 0)
                The column type information remains unchanged.
  Revised     - 1998.1116 (KSB) Wrote initial version.
---------------------------------------------------------------------------}
  VAR
    liCol     : WORD;
    liRow     : WORD;
  BEGIN
    FOR liRow := RowCount DOWNTO 2 DO BEGIN
      DeleteRow(liRow-1);
    END {FOR};

    FOR liCol := 1 TO ColCount DO BEGIN
      SetCell_AsString(liCol-1,0,'');
    END {FOR};
{}END {ClearData};




{}CONSTRUCTOR TDataStore.Create;
{+O
---------------------------------------------------------------------------
  Purpose     - Create the data store, it header and set the OnTypeChange
                link between the two, so that when the data type of a column
                header changes all the associated data values are converted
                to the appropriate data type.
---------------------------------------------------------------------------}
  BEGIN
    INHERITED Create;
    FiRow := 0;
    FiCol := 0;
    FcDelimt := DefaultDelimiter;
    FoNames  := TColumnDefs.Create;
    FoNames.OnTypeChange := ChangeColumnType;
{}END {Create};




{}PROCEDURE   TDataStore.DefineProperties(aoFiler: TFiler);
{+O
---------------------------------------------------------------------------
  Purpose     - Define psuedo properties that can be stored in the DFM.
---------------------------------------------------------------------------}
  BEGIN
    aoFiler.DefineProperty( 'RowCount', ReadRow, WriteRow, Count>0 );
    aoFiler.DefineProperty( 'ColCount', ReadCol, WriteCol, Count>0 );
    INHERITED DefineProperties(aoFiler);
{}END {DefineProperties};




{}PROCEDURE  TDataStore.DeleteRow( aiRow : WORD );
{+O
---------------------------------------------------------------------------
  Purpose     - Delete the indicated row in the data store.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    Delete( aiRow*FiCol );
    FOR li := 2 TO FiCol DO BEGIN
      Delete( aiRow*FiCol );
    END {FOR};

    Dec(FiRow);
{}END {DeleteRow};




{}DESTRUCTOR  TDataStore.Destroy;
{+O
---------------------------------------------------------------------------
  Purpose     - Release all previously allocated memory;
---------------------------------------------------------------------------}
  BEGIN
    FoNames.Free;
    INHERITED Destroy;
{}END {Destroy};




{}FUNCTION    TDataStore.GetCell(aiCol, aiRow: WORD): STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Get the untranslated cell contents.
---------------------------------------------------------------------------}
  BEGIN
    Result := Strings[aiRow*FiCol+aiCol];
{}END {GetCell};




{}FUNCTION    TDataStore.GetCell_AsString(aiCol, aiRow:WORD): STRING;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a readable string for the hex-coded stored data.
---------------------------------------------------------------------------}
{}{}FUNCTION HexToBoolStr(aiCol,aiRow:WORD):STRING;
    VAR
      li      : INTEGER;
      lsHex   : STRING;
    BEGIN
      lsHex := Cell[aiCol,aiRow];
      IF Length(lsHex) > 0 THEN BEGIN
        li := _4ByteHexStrToInt(lsHex);
        IF LongBOOL(li) THEN BEGIN
          result := 'TRUE'
        END ELSE BEGIN
          result := 'FALSE';
        END {IF};
      END ELSE BEGIN
        result := '';
      END {IF};
{}{}END {HexToBoolStr};


{}{}FUNCTION HexToFloat(aiCol,aiRow:WORD):DOUBLE;
    VAR
      lsHex   : STRING;
    BEGIN
      lsHex := Cell[aiCol,aiRow];
      IF Length(lsHex) > 0 THEN BEGIN
        result := _8ByteHexStrToFloat(lsHex);
      END ELSE BEGIN
        result := 0;
      END {IF};
{}{}END {HexToFloat};


{}{}FUNCTION HexToInt(aiCol,aiRow:WORD):INTEGER;
    VAR
      lsHex   : STRING;
    BEGIN
      lsHex := Cell[aiCol,aiRow];
      IF Length(lsHex) > 0 THEN BEGIN
        result := _4ByteHexStrToInt(lsHex);
      END ELSE BEGIN
        result := 0;
      END {IF};
{}{}END {HexToInt};


  BEGIN
    CASE Header.ColumnType[aiCol] OF
      ftBoolean : result := HexToBoolStr(aiCol,aiRow);
      ftCurrency: result := FloatToStrF(_8ByteHexStrToCurrency(Cell[aiCol,aiRow]),ffCurrency,20,2);
      ftDate    : result := FormatDateTime('dd/mmm/yyyy',HexToFloat(aiCol,aiRow));
      ftDateTime: result := FormatDateTime('dd/mmm/yyyy hh:mm:ss',HexToFloat(aiCol,aiRow));
      ftFloat   : result := FloatToStr(HexToFloat(aiCol,aiRow));
      ftInteger : result := IntToStr(HexToInt(aiCol,aiRow));
      ftSmallInt: result := IntToStr(HexToInt(aiCol,aiRow));
      ftString  : result := Cell[aiCol,aiRow];
      ftTime    : result := FormatDateTime('hh:mm:ss',HexToFloat(aiCol,aiRow));
      ftWord    : result := IntToStr(HexToInt(aiCol,aiRow));
    END {CASE};
{}END {GetCell_AsString};




{}FUNCTION    TDataStore.GetCell_Boolean (aiCol, aiRow: WORD): Boolean;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a boolean value for the contents of the cell.
---------------------------------------------------------------------------}
  BEGIN
    result := FALSE;
    IF Header.ColumnType[aiCol] = ftBoolean THEN BEGIN
      result := LongBOOL(GetCell_Integer(aiCol,aiRow));
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftBoolean);
    END {IF};
{}END {GetCell_Boolean};




{}FUNCTION    TDataStore.GetCell_Currency(aiCol, aiRow: WORD): Currency;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a currency value for the contents of the cell.
---------------------------------------------------------------------------}
  BEGIN
    result := 0;
    IF Header.ColumnType[aiCol] = ftCurrency THEN BEGIN
      result  := _8ByteHexStrToCurrency(Cell[aiCol,aiRow]);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftCurrency);
    END {IF};
{}END {GetCell_Currency};




{}FUNCTION    TDataStore.GetCell_DateTime(aiCol, aiRow: WORD): TDateTime;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a date/time value for the contents of the cell.
---------------------------------------------------------------------------}
  BEGIN
    result := 0;
    IF Header.ColumnType[aiCol] IN [ftDate,ftDateTime,ftTime] THEN BEGIN
      result  := _8ByteHexStrToFloat(Cell[aiCol,aiRow]);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftDateTime);
    END {IF};
{}END {GetCell_DateTime};




{}FUNCTION    TDataStore.GetCell_Float(aiCol, aiRow: WORD): Double;
{+O
---------------------------------------------------------------------------
  Purpose     - Return a floating point value for the contents of the cell.
---------------------------------------------------------------------------}
  BEGIN
    result := 0;
    IF Header.ColumnType[aiCol] = ftFloat THEN BEGIN
      result  := _8ByteHexStrToFloat(Cell[aiCol,aiRow]);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftFloat);
    END {IF};
{}END {GetCell_Float};




{}FUNCTION    TDataStore.GetCell_Integer(aiCol, aiRow: WORD): LongINT;
{+O
---------------------------------------------------------------------------
  Purpose     - Return an integer value for the contents of the cell.
---------------------------------------------------------------------------}
  BEGIN
    result := 0;
    IF Header.ColumnType[aiCol] IN [ftBoolean,ftInteger,ftSmallInt,ftWord] THEN BEGIN
      result  := _4ByteHexStrToInt(Cell[aiCol,aiRow]);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftInteger);
    END {IF};
{}END {GetCell_Integer};




{}FUNCTION    TDataStore.GetCObject(aiCol: WORD): TObject;
{+O
---------------------------------------------------------------------------
  Purpose     - Return the object for the "aiCol" column.
---------------------------------------------------------------------------}
  BEGIN
    Result := Objects[aiCol];
{}END {GetCObject};




{}FUNCTION    TDataStore.GetRObject(aiRow: WORD): TObject;
{+O
---------------------------------------------------------------------------
  Purpose     - Retrieve the object from the first column of the row.
---------------------------------------------------------------------------}
  BEGIN
    Result := Objects[aiRow*FiCol];
{}END {GetRObject};




{}FUNCTION    TDataStore.IndexOfObject(CONST aoObject:TObject):INTEGER;
{+O
---------------------------------------------------------------------------
  Purpose     - Return the row index for "aoObject"
---------------------------------------------------------------------------}
  BEGIN
    Result := INHERITED IndexOfObject(aoObject);
    IF Result > 0 THEN BEGIN
      Result := Result DIV FiCol;
    END {IF};
{}END {IndexOfObject};




{}PROCEDURE   TDataStore.InsertRow(aiRow:WORD; asStr:STRING; aoObj:TObject);
{+O
---------------------------------------------------------------------------
  Purpose     - Insert a new row at row "aiRow". The new column contains
                "asStr" & "aoObj" in column 0 of "aiRow" and "asStr" for
                every other column of that row.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
  BEGIN
    InsertObject( aiRow*FiCol, asStr, aoObj );
    FOR li := 2 TO FiCol DO BEGIN
      Insert( aiRow*FiCol+li-1, '' );
    END {FOR};

    Inc(FiRow);
{}END {InsertRow};




{}PROCEDURE TDataStore.LoadFromFile( CONST asFileName : STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Load the data store from the named file.
  Remarks     - Data columns are assumed to be separated by the "delimiting"
                character (usually a tab) stored in FcDelimit.
                Blank lines will terminate data loading unless FbBlankLines
                is true.
  Warning     - Partial lines are NOT guarenteed to be parsed correctly.
---------------------------------------------------------------------------}
{}{}FUNCTION  FetchDelimitedSubstring( VAR apStr : PChar ) : STRING;
    VAR
      lp1     : PChar;
    BEGIN
              {return a pointer to the first occurance of FcDelimit in apStr}
      IF apStr = NIL THEN BEGIN
        lp1 := NIL;
      END ELSE BEGIN
        lp1 := StrScan( apStr, FcDelimt );
      END {IF};
              {if a delimiter was found, then replace the delimiter with a
               #0 (effectively terminating the substring)}
      IF lp1 <> NIL THEN BEGIN
        lp1[0] := #0;
      END {IF};
              {return the substring before the delimiter as the function result
              (if the delimiter was found, the substring preceeding it is now
              terminated by #0. if no delimiter was found then the string is
              terminated by its normally occurring terminating #0)}
      Result := StrPas( apStr );
              {If a delimiter was found, the new pchar value is the substring
               that occurs after the (former) delimiter, otherwise we have
               reached the end of the pchar string...}
      IF lp1 <> NIL THEN
        apStr := lp1 + 1
      ELSE BEGIN
        apStr := NIL;
      END {IF};
{}{}END {FetchDelimitedSubstring};


  TYPE
    TStrBuf   = ARRAY [0..ciMaxRecLength] OF CHAR;
    PStrBuf   = ^TStrBuf;
  VAR
    li        : INTEGER;
    lTdata    : TextFile;
    lpRead    : PStrBuf;
    lpTemp    : PStrBuf;
  BEGIN
    AssignFile( lTdata, asFileName );
    Reset( lTdata );
    Clear;
    lpRead := New(PStrBuf);
    ReadLn( lTdata, lpRead^ );
    lpTemp := lpRead;
    WHILE lpTemp <> NIL DO BEGIN
      Add(FetchDelimitedSubstring( PChar(lpTemp) ));
    END {WHILE};

    FiCol := Count;
    FiRow := 1;
    WHILE NOT Eof( lTdata ) DO BEGIN
      ReadLn( lTdata, lpRead^ );
      lpTemp := lpRead;
      IF (lpTemp^[0]=#0) THEN BEGIN
        IF FbBlankLines THEN BEGIN
          FOR li := 1 TO FiCol DO BEGIN
            Add('');
          END {FOR};
        END ELSE BEGIN
          Break;
        END {IF}
      END ELSE BEGIN
        FOR li := 1 TO FiCol DO BEGIN
          Add(FetchDelimitedSubstring( PChar(lpTemp) ));
        END {FOR};
      END {IF};

      Inc( FiRow );
    END {WHILE};

    Dispose(lpRead );
    CloseFile( lTdata );
{}END {LoadFromFile};




{}PROCEDURE  TDataStore.PutCObject(aiCol: WORD; CONST aoObject: TObject);
{+O
---------------------------------------------------------------------------
  Purpose     - Store an object for column "aiCol"
---------------------------------------------------------------------------}
  BEGIN
    Objects[aiCol] := aoObject;
{}END {PutObject};




{}PROCEDURE   TDataStore.PutRObject(aiRow: WORD; CONST aoObject: TObject);
{+O
---------------------------------------------------------------------------
  Purpose     - Store an object in the first column of the row "aiRow".
---------------------------------------------------------------------------}
  BEGIN
    Objects[aiRow*FiCol] := aoObject;
{}END {PutRObject};




{}PROCEDURE   TDataStore.ReadCol(aoReader: TReader);
{+O
---------------------------------------------------------------------------
  Purpose     - Fetch the number of columns from the DFM.
---------------------------------------------------------------------------}
  BEGIN
    FiCol := aoReader.ReadInteger;
{}END {ReadCol};




{}PROCEDURE   TDataStore.ReadRow(aoReader: TReader);
{+O
---------------------------------------------------------------------------
  Purpose     - Fetch the number of rows from the DFM.
---------------------------------------------------------------------------}
  BEGIN
    FiRow := aoReader.ReadInteger;
{}END {ReadRow};




{}PROCEDURE   TDataStore.SaveToFile(CONST asFileName : STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Write the contents of the data store to the named file.
---------------------------------------------------------------------------}
  VAR
    li        : INTEGER;
    liJ       : INTEGER;
    lTdata    : TextFile;
  BEGIN
    AssignFile( lTdata, asFileName );
    Rewrite( lTdata );
    FOR li := 1 TO FiRow DO BEGIN
      Write( lTdata, Strings[(li-1)*FiCol] );
      FOR liJ := 2 TO FiCol DO BEGIN
        Write( lTdata, FcDelimt+Strings[(li-1)*FiCol+liJ-1] );
      END {FOR};

      WriteLn( lTdata );
    END {FOR};

    CloseFile( lTdata );
{}END {SaveToFile};




{}PROCEDURE   TDataStore.SetCell(aiCol, aiRow: WORD; CONST asValue: STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Directly store a string in the indicated cell
                (the string value "asValue" is not translated)
---------------------------------------------------------------------------}
  BEGIN
    Strings[aiRow*FiCol+aiCol] := asValue
{}END {SetCell};




{}PROCEDURE   TDataStore.SetCell_AsString(aiCol, aiRow:WORD; aValue:STRING);
{+O
---------------------------------------------------------------------------
  Purpose     - Convert the readable string to hex-coded data (unless its
                target column is "ftString") and store it.
  Revised     - 1998.1106 (KSB) Corrected Time conversion.
---------------------------------------------------------------------------}
{}{}FUNCTION StrToHexBool:STRING;
    BEGIN
      aValue := UpperCase(aValue);
      IF (aValue = 'TRUE') OR (aValue = 'YES') THEN BEGIN
        result := IntTo4ByteHexStr(Ord(TRUE));
      END ELSE
      IF (aValue = 'FALSE') OR (aValue = 'NO') THEN BEGIN
        result := IntTo4ByteHexStr(Ord(FALSE));
      END ELSE BEGIN
        result := '';
      END {IF};
{}{}END {StrToHexBool};


{}{}FUNCTION StrToHexCurrency:STRING;
    VAR
      lfDollar: CURRENCY;
      ls      : STRING;
    BEGIN
      ls := UpperCase(aValue);
      IF ls = '' THEN BEGIN
        lfDollar := 0;
      END ELSE BEGIN
        lfDollar := StrToCurrency(ls);
      END {IF};

      result := CurrencyTo8ByteHexStr(lfDollar);
{}{}END {StrToHexCurrency};


{}{}FUNCTION StrToHexDate(aeType:TFieldType):STRING;
    VAR
      lfDT    : TDateTime;
    BEGIN
      IF aValue = '' THEN BEGIN
        lfDT := 0;
      END ELSE BEGIN
        CASE aeType OF
          ftDate,
          ftDateTime : lfDT := KSSDates.KSS_StrToDateTime(aValue);
          ftTime     : lfDT := KSSDates.KSS_StrToTime(aValue);
          ELSE         lfDT := 0;
        END {CASE};
      END {IF};

      result := DoubleTo8ByteHexStr(lfDT);
{}{}END {StrToHexDate};


{}{}FUNCTION StrToHexFloat:STRING;
    VAR
      lfD     : DOUBLE;
    BEGIN
      IF aValue = '' THEN BEGIN
        lfD := 0.0;
      END ELSE BEGIN
        lfD := SysUtils.StrToFloat(aValue);
      END {IF};

      result := DoubleTo8ByteHexStr(lfD);
{}{}END {StrToHexFloat};


{}{}FUNCTION StrToHexInt:STRING;
    VAR
      li      : INTEGER;
    BEGIN
      li := SysUtils.StrToIntDef(aValue,0);
      CASE Header.ColumnType[aiCol] OF
        ftWord     : BEGIN
          IF (li < 0) OR (li > 65535) THEN BEGIN
            BadWordValue(li);
          END {IF};
        END {ftWord};

        ftSmallInt : BEGIN
          IF (li < -32768) OR (li > 32767) THEN BEGIN
            BadSmallIntValue(li);
          END {IF};
        END {ftSmallInt};

        ftInteger  :;
      END {CASE};

      result := IntTo4ByteHexStr(li);
{}{}END {StrToHexInt};


  BEGIN
    CASE Header.ColumnType[aiCol] OF
      ftBoolean : Cell[aiCol,aiRow] := StrToHexBool;
      ftCurrency: Cell[aiCol,aiRow] := StrToHexCurrency;
      ftDate    : Cell[aiCol,aiRow] := StrToHexDate(Header.ColumnType[aiCol]);
      ftDateTime: Cell[aiCol,aiRow] := StrToHexDate(Header.ColumnType[aiCol]);
      ftFloat   : Cell[aiCol,aiRow] := StrToHexFloat;
      ftInteger : Cell[aiCol,aiRow] := StrToHexInt;
      ftSmallInt: Cell[aiCol,aiRow] := StrToHexInt;
      ftString  : Cell[aiCol,aiRow] := aValue;
      ftTime    : Cell[aiCol,aiRow] := StrToHexDate(Header.ColumnType[aiCol]);
      ftWord    : Cell[aiCol,aiRow] := StrToHexInt;
    END {CASE};
{}END {SetCell_AsString};




{}PROCEDURE   TDataStore.SetCell_Boolean (aiCol, aiRow: WORD; aValue:BOOLEAN);
{+O
---------------------------------------------------------------------------
  Purpose     - Store a boolean value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    SetCell_Integer(aiCol,aiRow,Ord(aValue));
{}END {SetCell_Boolean};




{}PROCEDURE   TDataStore.SetCell_Currency(aiCol, aiRow: WORD; aValue:CURRENCY);
{+O
---------------------------------------------------------------------------
  Purpose     - Store a currency value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF Header.ColumnType[aiCol] IN [ftCurrency] THEN BEGIN
      Cell[aiCol,aiRow] := CurrencyTo8ByteHexStr(aValue);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftCurrency);
    END {IF};
{}END {SetCell_Currency};




{}PROCEDURE   TDataStore.SetCell_Date(aiCol, aiRow: WORD; aValue:TDateTime);
{+O
---------------------------------------------------------------------------
  Purpose     - Store just the date portion of the date/time value.
                Store the date value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    SetCell_DateTime(aiCol,aiRow,Int(aValue));
{}END {SetCell_Date};




{}PROCEDURE   TDataStore.SetCell_DateTime(aiCol, aiRow: WORD; aValue:TDateTime);
{+O
---------------------------------------------------------------------------
  Purpose     - Store the date/time value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF Header.ColumnType[aiCol] IN [ftDate,ftDateTime,ftTime] THEN BEGIN
      Cell[aiCol,aiRow] := DoubleTo8ByteHexStr(aValue);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftDateTime);
    END {IF};
{}END {SetCell_DateTime};




{}PROCEDURE   TDataStore.SetCell_Float   (aiCol, aiRow: WORD; aValue:DOUBLE);
{+O
---------------------------------------------------------------------------
  Purpose     - Store the floating point value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF Header.ColumnType[aiCol] IN [ftFloat] THEN BEGIN
      Cell[aiCol,aiRow] := DoubleTo8ByteHexStr(aValue);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftFloat);
    END {IF};
{}END {SetCell_Float};




{}PROCEDURE   TDataStore.SetCell_Integer (aiCol, aiRow: WORD; aValue:LongINT);
{+O
---------------------------------------------------------------------------
  Purpose     - Store the integer value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF Header.ColumnType[aiCol] IN [ftBoolean,ftSmallINT,ftInteger,ftWord] THEN BEGIN
      Cell[aiCol,aiRow] := IntTo4ByteHexStr(aValue);
    END ELSE BEGIN
      ColumnTypeIsWrong(aiCol,ftInteger);
    END {IF};
{}END {SetCell_Integer};




{}PROCEDURE   TDataStore.SetCell_SmallInt(aiCol, aiRow: WORD; aValue:LongINT);
{+O
---------------------------------------------------------------------------
  Purpose     - Store the small integer value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF (aValue < -32768) OR (aValue > 32767) THEN BEGIN
      RAISE ERangeError.Create(IntToStr(aValue)+' is not a SmallINT value.');
    END ELSE BEGIN
      SetCell_Integer(aiCol,aiRow,aValue);
    END {IF};
{}END {SetCell_SmallInt};




{}PROCEDURE   TDataStore.SetCell_Time(aiCol, aiRow: WORD; aValue:TDateTime);
{+O
---------------------------------------------------------------------------
  Purpose     - Store just the time portion of the value.
                Store the time value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    SetCell_DateTime(aiCol,aiRow,Frac(aValue));
{}END {SetCell_Time};




{}PROCEDURE   TDataStore.SetCell_Word(aiCol, aiRow: WORD; aValue:LongINT);
{+O
---------------------------------------------------------------------------
  Purpose     - Store the word value in the indicated cell
                (the "aValue" is translated to a hex-string before storing)
---------------------------------------------------------------------------}
  BEGIN
    IF (aValue < 0) OR (aValue > 65535) THEN BEGIN
      RAISE ERangeError.Create(IntToStr(aValue)+' is not a WORD value.');
    END ELSE BEGIN
      SetCell_Word(aiCol,aiRow,aValue);
    END {IF};
{}END {SetCell_Word};




{}PROCEDURE   TDataStore.SetStoreColumns(aiColumns:WORD);
{+O
---------------------------------------------------------------------------
  Purpose     - Set the initial number of data store columns.
---------------------------------------------------------------------------}
  BEGIN
    IF FiCol = 0 THEN BEGIN
      FiCol := aiColumns;
      InsertRow( 0, '', POINTER(0) );
    END {IF};
{}END {SetStoreColumns};




{}PROCEDURE   TDataStore.WriteCol( aoWriter: TWriter);
{+O
---------------------------------------------------------------------------
  Purpose     - Save the number of columns to the DFM.
---------------------------------------------------------------------------}
  BEGIN
    aoWriter.WriteInteger( FiCol );
{}END {WriteCol};




{}PROCEDURE   TDataStore.WriteRow( aoWriter: TWriter);
{+O
---------------------------------------------------------------------------
  Purpose     - Save the number of rows to the DFM.
---------------------------------------------------------------------------}
  BEGIN
    aoWriter.WriteInteger( FiRow );
{}END {WriteRow};


END {KSSCnvrt}.
